# -*- coding: utf-8 -*-
#
# Author: PaperCube aka 张老师
# Github: github.com/PaperCube
# Date:   20230421

import os
import subprocess
import sys
import time
from types import ModuleType
from typing import List, Literal, Union


class DaemonableProcess:
    def __init__(
        self,
        *,
        target: Union[str, ModuleType],
        args: List = [],
        mode: Literal["attached", "detached"],
        redirect_stdout=None,
        redirect_stderr=subprocess.STDOUT,
        use_os_spawn_for_daemon=False,
        redirect_stdin=subprocess.DEVNULL,
        env_append=None,
    ):
        self.target = target
        self.args = args
        self._mode = mode
        self.use_os_spawn_for_daemon = use_os_spawn_for_daemon
        self.redirect_stdin = redirect_stdin
        # self.__is_daemon = is_daemon
        # self.__is_detached = is_detached
        self.redirect_stdout = redirect_stdout
        self.redirect_stderr = redirect_stderr
        self.env = dict(os.environ)
        if env_append is not None:
            self.env.update(env_append)
        # self.__no_new_window = no_new_window

    @property
    def is_daemon(self):
        return self._mode == "daemon"

    @property
    def mode(self):
        return self._mode

    # @property
    # def is_detached(self):
    #     return self.__is_detached

    def start(self, shell=False, path=None):
        if self.use_os_spawn_for_daemon:
            raise NotImplementedError()
        else:
            # use subprocess
            command_line = (
                [sys.executable, "-m", self.target.__name__, *self.args]
                if isinstance(self.target, ModuleType)
                else [self.target, *self.args]
            )
            path = path or os.getcwd()
            is_ms_windows = "win32" in sys.platform or "cygwin" in sys.platform
            if is_ms_windows:
                # windows + subprocess
                creationflags = {
                    "attached": 0,
                    "shared": 0,
                    "detached": subprocess.CREATE_NO_WINDOW,  # type: ignore (only for windows)
                }[self._mode]

                popen = subprocess.Popen(
                    command_line,
                    creationflags=creationflags,
                    stdout=self.redirect_stdout,
                    stderr=self.redirect_stderr,
                    stdin=self.redirect_stdin,
                    env=self.env,
                    cwd=path,
                    shell=shell,
                )
                # print(popen)

                # return_code = popen.wait()
                # print('return_code', return_code)
            else:
                # posix + subprocess
                popen = subprocess.Popen(
                    command_line,
                    stdin=self.redirect_stdin,
                    stdout=self.redirect_stdout,
                    stderr=self.redirect_stderr,
                    env=self.env,
                    start_new_session=self._mode == "detached",
                    cwd=path,
                    shell=shell,
                )

                if self._mode == "attached":
                    import atexit

                    atexit.register(lambda: popen.terminate())

                # print(popen)

            return popen

    def terminate(self):
        pass


def main():
    DaemonableProcess(
        target="ppcdaemontest.daemon_worker",
        args=["1"],
        mode="attached",
    ).start()
    time.sleep(3)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231204

from typing import Dict, List

from neetbox._protocol import *
from neetbox.logging import Logger

from ....db.project import ProjectDB

logger = Logger("Project Bridge", skip_writers_names=["ws"])


class Bridge:
    """Server works as a bridge between clients and frontends. Bridge connects one client and multiple frontend. Since client means a running project, a bridge represents a running(connected) or not running(not connected, loaded from history database) project.
    You can:
    - send websocket message to clients or frontends connected to specific bridge
    - read and write history db(sqlite3) via bridge
    HTTP server should serves information via bridges.

            Client  ┌──--> Frontend
                |   |
                ↓   ↓
    Client<--> Server--> Frontend
                ↑   ↑
                │   │
            Client  └──--> Frontend
    """

    # static
    _id2bridge: Dict[str, "Bridge"] = {}  # manage connections using project id

    # instance vars
    project_id: str
    status: dict
    cli_ws_dict: dict  # { run_id : ws_client}
    web_ws_list: dict  # since web do not have run id, use list instead of dict
    historyDB: ProjectDB

    def __new__(cls, project_id: str, **kwargs) -> None:
        """Create Bridge of project id, return the old one if already exist
        Args:
            project_id (str): project id
        Returns:
            Bridge: Bridge of given project id
        """
        if project_id not in cls._id2bridge:
            new_bridge = super().__new__(cls)
            new_bridge.project_id = project_id
            new_bridge.cli_ws_dict: dict = {}  # cli ws sid
            new_bridge.web_ws_list: list = (
                []
            )  # frontend ws sids. client data should be able to be shown on multiple frontend
            flag_auto_load_db = kwargs["auto_load_db"] if "auto_load_db" in kwargs else True
            new_bridge.historyDB = ProjectDB.get_db_of_id(project_id) if flag_auto_load_db else None
            cls._id2bridge[project_id] = new_bridge
            logger.info(f"created new Bridge for project id '{project_id}'")
        return cls._id2bridge[project_id]

    def __del__(self):  # on delete
        logger.info(f"bridge project id {self.project_id} handling on delete...")
        if 0 == len(self.get_run_ids()):  # if there is no active run id
            self.historyDB.delete()
            del self.historyDB  # delete history db
        logger.info(f"bridge of project id {self.project_id} deleted.")

    @classmethod
    def items(cls):
        return cls._id2bridge.items()

    @classmethod
    def has(cls, project_id: str):
        return project_id in cls._id2bridge

    @classmethod
    def of_id(cls, project_id: str) -> "Bridge":
        """get Bridge of project id if exist. note that this class method is different fromBridge.__new__, which do not create new Bridge if given id not exist, it returns Noneinstead.
        Args:
            project_id (str): project id
        Returns:
            Bridge: Bridge if id exist, otherwise None
        """
        bridge = cls._id2bridge[project_id] if cls.has(project_id) else None
        return bridge

    @classmethod
    def from_db(cls, db: ProjectDB) -> "Bridge":
        project_id = db.fetch_db_project_id()
        bridge = Bridge(project_id, auto_load_db=False)
        if bridge.historyDB is not None:
            logger.warn(f"overwriting db of '{project_id}'")
        bridge.historyDB = db
        return bridge

    @classmethod
    def load_histories(cls):
        db_list = ProjectDB.get_db_list()
        logger.log(f"found {len(db_list)} history db.")
        for _, history_db in db_list:
            cls.from_db(history_db)

    async def ws_send_to_frontends(self, message: EventMsg):
        for ws_client in self.web_ws_list:
            try:
                await ws_client.ws.send_json(data=message.json)
            except Exception as e:
                logger.err(e)
        return

    async def ws_send_to_client(self, message: EventMsg, run_id: str = None):
        run_id = run_id or message.run_id
        ws_client = self.cli_ws_dict[run_id]
        try:
            await ws_client.ws.send_json(message.json)
        except Exception as e:
            logger.err(e)
        return

    def set_status(self, run_id: str, series: str, value: dict):
        self.historyDB.set_status(run_id=run_id, series=series, json_data=value)

    def get_status(self, run_id: str = None, series: str = None):
        status = self.historyDB.get_status(run_id=run_id, series=series)
        if run_id:
            status = status.get(run_id, {})
        if series:
            status = status.get(series, {})
        return status

    def is_online(self, run_id: str = None):
        if run_id:  # if checking client run with specific run id
            return run_id in self.cli_ws_dict
        return len(self.cli_ws_dict.keys()) != 0

    def get_series_of(self, table_name, run_id=None):
        return self.historyDB.get_series_of_table(table_name=table_name, run_id=run_id)

    def get_run_ids(self):
        info_run_ids = self.historyDB.get_run_ids()
        for info_run_id in info_run_ids:
            info_run_id["online"] = info_run_id[RUN_ID_KEY] in self.cli_ws_dict
        return info_run_ids

    def save_json_to_history(
        self, table_name, json_data, series=None, run_id=None, timestamp=None, num_row_limit=-1
    ):
        lastrowid = Bridge.of_id(self.project_id).historyDB.write_json(
            table_name=table_name,
            json_data=json_data,
            series=series,
            run_id=run_id,
            timestamp=timestamp,
            num_row_limit=num_row_limit,
        )
        return lastrowid

    def read_json_from_history(self, table_name, condition):
        return self.historyDB.read_json(table_name=table_name, condition=condition)

    def save_blob_to_history(
        self,
        table_name,
        meta_data,
        blob_data,
        series=None,
        run_id=None,
        timestamp=None,
        num_row_limit=-1,
    ):
        lastrowid = Bridge.of_id(self.project_id).historyDB.write_blob(
            table_name=table_name,
            meta_data=meta_data,
            blob_data=blob_data,
            series=series,
            run_id=run_id,
            timestamp=timestamp,
            num_row_limit=num_row_limit,
        )
        return lastrowid

    def read_blob_from_history(self, table_name, condition, meta_only: bool):
        return self.historyDB.read_blob(table_name, condition=condition, meta_only=meta_only)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231211

from neetbox._protocol import *
from neetbox.utils.x2numpy import *

from .._client import connection

# ===================== PLOTTING things ===================== #


def add_scalar(name: str, x: Union[int, float], y: Union[int, float]):
    """send a scalar to frontend display

    Args:
        name (str): name of the image, used in frontend display
        x (Union[int, float]): x
        y (Union[int, float]): y
    """
    # send
    connection.ws_send(event_type=EVENT_TYPE_NAME_SCALAR, series=name, payload={"x": x, "y": y})


# ===================== HYPERPARAM things ===================== #


def add_hyperparams(hparam: dict, name: str = None):
    """add/set hyperparams to current run, the added hyperparams will show in frontend

    Args:
        hparam (dict): hyperparams
        name (str, optional): name of hyperparams. Defaults to None.
    """
    assert isinstance(hparam, dict)
    hparam = {name: hparam} if name else hparam
    connection.ws_send(event_type=EVENT_TYPE_NAME_HPARAMS, series=name, payload=hparam)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231224

import os
import pathlib
import sys


def get_user_config_directory():
    """Returns a platform-specific root directory for user config settings.
    On Windows, prefer %LOCALAPPDATA%, then %APPDATA%, since we can expect the AppData directories to be ACLed to be visible only to the user and admin users (https://stackoverflow.com/a/7617601/1179226). If neither is set, return None instead of falling back to something that may be world-readable.
    """
    if os.name == "nt":
        appdata = os.getenv("LOCALAPPDATA")
        if appdata:
            return appdata
        appdata = os.getenv("APPDATA")
        if appdata:
            return appdata
        return None
    # On non-windows, use XDG_CONFIG_HOME if set, else default to ~/.config.
    xdg_config_home = os.getenv("XDG_CONFIG_HOME")
    if xdg_config_home:
        return xdg_config_home
    return os.path.join(os.path.expanduser("~"), ".config")


def get_create_neetbox_config_directory():
    path = os.path.join(get_user_config_directory(), "neetbox")
    if not os.path.exists(path):
        os.makedirs(path)
    assert os.path.isdir(
        path
    ), f"Fialed to create neetbox config directory {path}, please check your permission or create it manually."
    return path


def get_user_app_data_directory():
    """
    Returns a parent directory path
    where persistent application data can be stored.

    - linux: ~/.local/share
    - macOS: ~/Library/Application Support
    - windows: C:/Users/<USER>/AppData/Roaming
    """

    home = pathlib.Path.home()

    if sys.platform == "win32":
        app_data_path = home / "AppData/Roaming"
    elif sys.platform == "linux":
        app_data_path = home / ".local/share"
    elif sys.platform == "darwin":
        app_data_path = home / "Library/Application Support"

    return str(app_data_path)


def get_create_neetbox_data_directory():
    path = os.path.join(get_user_app_data_directory(), "neetbox")
    if not os.path.exists(path):
        os.makedirs(path)
    assert os.path.isdir(
        path
    ), f"Fialed to create neetbox data directory {path}, please check your permission or create it manually."
    return path


def get_file_size_in_bytes(file_path):
    return os.path.getsize(file_path)


def get_folder_size_in_bytes(folder_path, skip_symbolic_link=True):
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(folder_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            if os.path.islink(fp) and skip_symbolic_link:
                continue  # Skip if it is symbolic link
            total_size += os.path.getsize(fp)
    return total_size



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240118

import time
from threading import Thread
from typing import Callable, List

import psutil

from neetbox._protocol import *
from neetbox.config.user import get as get_global_config
from neetbox.utils.localstorage import get_folder_size_in_bytes
from neetbox.utils.mvc import Singleton


class CpuStatus:
    def __init__(self, id=-1, percentage=0.0, frequency=0.0):
        self.id = id
        self.percentage = percentage
        self.frequency = frequency

    def __str__(self) -> str:
        return f"CPU({self.id}), {(self.percentage)}%, {(self.frequency)}Mhz"

    @property
    def json(self):
        return {"id": self.id, "percentage": self.percentage, "frequency": self.frequency}


class CpuStatistics:
    def __init__(self, ctx_switches, interrupts, soft_interrupts, syscalls) -> None:
        self.ctx_switches = ctx_switches
        self.interrupts = interrupts
        self.soft_interrupts = soft_interrupts
        self.syscalls = syscalls

    @property
    def json(self):
        return {
            "ctxSwitches": self.ctx_switches,
            "interrupts": self.interrupts,
            "softInterrupts": self.soft_interrupts,
            "syscalls": self.syscalls,
        }

    def __str__(self) -> str:
        result = ""
        for k, v in self.json.items():
            result += f"{k}:\t{v}"
        return result


class MemoryStatus:
    def __init__(self, total, available, used, free) -> None:
        self.total = total
        self.available = available
        self.used = used
        self.free = free

    @property
    def json(self):
        return {
            "total": self.total,
            "available": self.available,
            "used": self.used,
            "free": self.free,
        }

    def __str__(self) -> str:
        result = ""
        for k, v in self.json.items():
            result += f"{k}:\t{v}"
        return result


class Hardware(metaclass=Singleton):
    def add_on_update_callback(self, func: Callable):
        self._on_update_call_backs.append(func)

    # properties
    _cpus: List[CpuStatus]
    _cpu_statistics: CpuStatistics
    _memory: MemoryStatus

    def __init__(self) -> None:
        self._cpus = [CpuStatus() for _ in range(psutil.cpu_count(logical=True))]
        self._cpu_statistics = CpuStatistics(*psutil.cpu_stats())
        virtual_memory = psutil.virtual_memory()
        self._memory = MemoryStatus(
            total=virtual_memory[0] / 1e6,
            available=virtual_memory[1] / 1e6,
            used=virtual_memory[3] / 1e6,
            free=virtual_memory[4] / 1e6,
        )

        def watch_thread():
            while True:
                # update cpu usage
                cpu_percent = psutil.cpu_percent(percpu=True)
                cpu_freq = psutil.cpu_freq(percpu=True)
                if len(cpu_freq) == 1:
                    cpu_freq = cpu_freq * len(cpu_percent)
                for index in range(len(cpu_percent)):
                    self._cpus[index] = CpuStatus(
                        id=index,
                        percentage=cpu_percent[index],
                        frequency=cpu_freq[index],
                    )
                # update memory usage
                virtual_memory = psutil.virtual_memory()
                self._memory = MemoryStatus(
                    total=virtual_memory[0] / 1e6,
                    available=virtual_memory[1] / 1e6,
                    used=virtual_memory[3] / 1e6,
                    free=virtual_memory[4] / 1e6,
                )
                self._cpu_statistics = CpuStatistics(*psutil.cpu_stats())
                time.sleep(1)

        Thread(target=watch_thread, daemon=True).start()

    @property
    def cpus(self):
        return self._cpus

    @property
    def cpu_statistics(self):
        return self._cpu_statistics

    @property
    def memory(self):
        return self._memory

    @property
    def json(self):
        return {
            "cpus": [cpu.json for cpu in self._cpus],
            "cpustat": self._cpu_statistics.json,
            "ram": self._memory.json,
        }


hardware = Hardware()  # singleton


class Disks(metaclass=Singleton):
    def __init__(self) -> None:
        def watch_thread():
            while True:
                time.sleep(30)

        Thread(target=watch_thread, daemon=True).start()

    @property
    def json(self):
        """Docstring.

        Returns:
            TYPE: Description
        """
        try:
            neetbox_data_dir = get_global_config("vault")
            neetbox_storage_in_bytes = get_folder_size_in_bytes(neetbox_data_dir)
            drive = psutil.disk_usage(neetbox_data_dir)
            return {
                "total": drive.total / 1e6,
                "used": drive.used / 1e6,
                "neetbox": neetbox_storage_in_bytes / 1e6,
                "free": drive.free / 1e6,
            }
        except Exception as e:
            return {ERROR_KEY: f"failed to get disk usage cause {e}"}


disks = Disks()  # singleton



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230414

import setproctitle
import uvicorn
from rich.console import Console
from rich.table import Table

from neetbox._protocol import *

console = Console()


def server_process(cfg, debug=False):
    setproctitle.setproctitle("NEETBOX SERVER")
    from neetbox.logging import Logger

    from .fastapi import serverapp

    logger = Logger("SERVER LAUNCHER", skip_writers_names=["ws"])

    port = cfg["port"]
    logger.log(f"launching fastapi server on http://0.0.0.0:{port}")
    uvicorn_log_level = "info" if debug else "critical"

    table = Table(title="Server Routes")
    table.add_column("name", style="green", no_wrap=True)
    table.add_column("path")
    for name, path in [(route.name, route.path) for route in serverapp.routes]:
        table.add_row(name, path)
    console.print(table)
    uvicorn.run(serverapp, host="0.0.0.0", port=port, log_level=uvicorn_log_level)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240116

import json
from abc import ABC, abstractmethod, abstractproperty
from enum import Enum
from typing import Dict, Tuple, Union


class FetchType(str, Enum):
    ALL = "all"
    ONE = "one"
    MANY = "many"


class SortType(str, Enum):
    ASC = "ASC"
    DESC = "DESC"


class ManageableDB(ABC):
    @abstractproperty
    def size(self):
        """get local storage usage in bytes"""
        ...

    @abstractmethod
    def close(self):
        ...

    @abstractmethod
    def delete(self):
        """handle delete, a typical behavior is to close connection and remove files"""
        ...



import asyncio
import os
import subprocess
from threading import Thread

from neetbox.config.project import _get_module_level_config
from neetbox.utils import DaemonableProcess
from neetbox.utils.mvc import Singleton


class ShellAgent(metaclass=Singleton):
    def __init__(self) -> None:
        shell_exec = "cmd.exe" if os.name == "nt" else "/bin/bash"
        self.daemonable = DaemonableProcess(
            command_line=shell_exec,
            redirect_stdin=subprocess.PIPE,
            redirect_stdout=subprocess.PIPE,
            redirect_stderr=subprocess.PIPE,
        )

    def _send_to_ws(self):
        pass

    def _send_to_stdin(self, string, flush=True):
        self.proc.redirect_stdin.write(string.encode())
        if flush:
            self.proc.redirect_stdin.flush()

    async def read_stdout(self):
        while True:
            output = self.proc.stdout.read(1)  # read stdio nonblocking
            if output:
                await self._send_to_ws(output.decode())

    async def start(self, path):
        path = path or os.getcwd()
        self.proc = self.daemonable.start(shell=True, path=path)

        await asyncio.gather(self.read_stdout())


# shell = ShellAgent()

if __name__ == "__main__":
    pass



import numpy as np

def check_nan(array):
    tmp = np.sum(array)
    if np.isnan(tmp) or np.isinf(tmp):
        pass # todo logger.warning('NaN or Inf found in input tensor.')
    return array


def make_np(x):
    if isinstance(x, list):
        return check_nan(np.array(x))
    if isinstance(x, np.ndarray):
        return check_nan(x)
    if isinstance(x, str):  # Caffe2 will pass name of blob(s) to fetch
        return check_nan(prepare_caffe2(x))
    if np.isscalar(x):
        return check_nan(np.array([x]))
    if 'torch' in str(type(x)):
        return check_nan(prepare_pytorch(x))
    if 'chainer' in str(type(x)):
        return check_nan(prepare_chainer(x))
    if 'mxnet' in str(type(x)):
        return check_nan(prepare_mxnet(x))
    if 'jax' in str(type(x)):
        return check_nan(np.array(x))
    raise NotImplementedError(
        'Got {}, but expected numpy array or torch tensor.'.format(type(x)))


def prepare_pytorch(x):
    import torch
    if isinstance(x, torch.autograd.Variable):
        x = x.data
    x = x.cpu().numpy()
    return x


def prepare_theano(x):
    import theano
    pass


def prepare_caffe2(x):
    from caffe2.python import workspace
    x = workspace.FetchBlob(x)
    return x


def prepare_mxnet(x):
    x = x.asnumpy()
    return x


def prepare_chainer(x):
    import chainer
    x = chainer.cuda.to_cpu(x.data)
    return x



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230417

import getpass
import platform as _platform

from neetbox.extension import on_workspace_loaded
from neetbox.utils.mvc import Singleton


class PlatformInfo(metaclass=Singleton):
    def __init__(self):
        self._username = getpass.getuser()
        self._machine = _platform.machine()
        self._processor = "unknown" if len(_platform.processor()) == 0 else _platform.processor()
        self._os_name = _platform.system()
        self._os_release = _platform.version()
        self._architecture = _platform.architecture()
        self._python_version = _platform.python_version()
        self._python_build = _platform.python_build()

    @property
    def username(self):
        return self._username

    @property
    def machine(self):
        return self._machine

    @property
    def processor(self):
        return self._processor

    @property
    def os_name(self):
        return self._os_name

    @property
    def os_release(self):
        return self._os_release

    @property
    def architecture(self):
        return self._architecture

    @property
    def python_version(self):
        return self._python_version

    @property
    def python_build(self):
        return self._python_build

    @property
    def json(self):
        return {
            "username": self._username,
            "machine": self._machine,
            "processor": self._processor,
            "os_name": self._os_name,
            "os_release": self._os_release,
            "architecture": self._architecture,
            "python_version": self._python_version,
            "python_build": self._python_build,
        }

    def __str__(self) -> str:
        result = ""
        for k, v in self.json.items():
            result += f"{k}:\t{v}"
        return result


platform = PlatformInfo()


# watch updates in daemon
@on_workspace_loaded(name="show-platform-information")
def load_send_platform_info():
    from neetbox._protocol import (
        EVENT_TYPE_NAME_HANDSHAKE,
        EVENT_TYPE_NAME_STATUS,
        EventMsg,
    )
    from neetbox.client import connection

    @connection.ws_subscribe(event_type_name=EVENT_TYPE_NAME_HANDSHAKE)
    def ws_send_platform_info(message: EventMsg):
        connection.ws_send(
            event_type=EVENT_TYPE_NAME_STATUS, series="platform", payload=platform.json
        )



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231201

import atexit
from collections import defaultdict
from typing import Dict

from neetbox.logging import Logger
from neetbox.utils.framing import get_caller_info_traceback
from neetbox.utils.mvc import Singleton

from .abc import ManageableDB

logger = Logger("DB Conn Manager", skip_writers_names=["ws"])

ModuleNameType = str
DbIdType = str


class DBConnectionManager(metaclass=Singleton):
    _POOL: Dict[ModuleNameType, Dict[DbIdType, ManageableDB]] = defaultdict(dict)

    @property
    def current(self):
        identity = get_caller_info_traceback(stack_offset=2)
        return self._POOL[identity.module_name]


manager = DBConnectionManager()


def clear_dbc_on_exit():
    logger.info("process exiting, cleaning up...")
    for module_name, conn_dict in manager._POOL.items():
        logger.info(f"closing db connection for module {module_name}:")
        dbc: ManageableDB
        for _, dbc in conn_dict.items():
            logger.info(f"=> closing {dbc}")
            try:
                dbc.close()
            except Exception as e:
                logger.err(RuntimeError(f"failed to close db connection {dbc}, {e}"))


atexit.register(clear_dbc_on_exit)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231203

import os
from uuid import uuid4

import toml

from neetbox._protocol import MACHINE_ID_KEY
from neetbox.utils.localstorage import (
    get_create_neetbox_config_directory,
    get_create_neetbox_data_directory,
)
from neetbox.utils.massive import check_read_toml

_GLOBAL_CONFIG = {
    MACHINE_ID_KEY: str(uuid4()),
    "vault": get_create_neetbox_data_directory(),
}

_GLOBAL_CONFIG_FILE_NAME = f"neetbox.global.toml"


def overwrite_create_local(config: dict):
    neetbox_config_dir = get_create_neetbox_config_directory()
    config_file_path = os.path.join(neetbox_config_dir, _GLOBAL_CONFIG_FILE_NAME)
    with open(config_file_path, "w+") as config_file:
        toml.dump(config, config_file)


def read_create_local():
    global _GLOBAL_CONFIG
    neetbox_config_dir = get_create_neetbox_config_directory()
    config_file_path = os.path.join(neetbox_config_dir, _GLOBAL_CONFIG_FILE_NAME)
    if not os.path.exists(config_file_path):  # config not exist, try to create
        overwrite_create_local(_GLOBAL_CONFIG)
    # read local file
    user_cfg = check_read_toml(config_file_path)
    assert user_cfg
    _GLOBAL_CONFIG.update(user_cfg)


def set(key, value):
    global _GLOBAL_CONFIG
    read_create_local()
    _GLOBAL_CONFIG[key] = value
    overwrite_create_local(_GLOBAL_CONFIG)


def get(key=None):
    read_create_local()
    if key is None:
        return _GLOBAL_CONFIG.copy()
    return _GLOBAL_CONFIG.get(key, None)



import numpy as np


def get_gaussian_kernel(k=3, mu=0, sigma=1, normalize=True):
    # compute 1 dimension gaussian
    gaussian_1D = np.linspace(-1, 1, k)
    # compute a grid distance from center
    x, y = np.meshgrid(gaussian_1D, gaussian_1D)
    distance = (x**2 + y**2) ** 0.5

    # compute the 2 dimension gaussian
    gaussian_2D = np.exp(-((distance - mu) ** 2) / (2 * sigma**2))
    gaussian_2D = gaussian_2D / (2 * np.pi * sigma**2)

    # normalize part (mathematically)
    if normalize:
        gaussian_2D = gaussian_2D / np.sum(gaussian_2D)
    return gaussian_2D


def get_sobel_kernel(k=3):
    # get range
    range = np.linspace(-(k // 2), k // 2, k)
    # compute a grid the numerator and the axis-distances
    x, y = np.meshgrid(range, range)
    sobel_2D_numerator = x
    sobel_2D_denominator = x**2 + y**2
    sobel_2D_denominator[:, k // 2] = 1  # avoid division by zero
    sobel_2D = sobel_2D_numerator / sobel_2D_denominator
    return sobel_2D


def get_thin_kernels(start=0, end=360, step=45):
    import cv2

    k_thin = 3  # actual size of the directional kernel
    # increase for a while to avoid interpolation when rotating
    k_increased = k_thin + 2

    # get 0° angle directional kernel
    thin_kernel_0 = np.zeros((k_increased, k_increased))
    thin_kernel_0[k_increased // 2, k_increased // 2] = 1
    thin_kernel_0[k_increased // 2, k_increased // 2 + 1 :] = -1

    # rotate the 0° angle directional kernel to get the other ones
    thin_kernels = []
    for angle in range(start, end, step):
        (h, w) = thin_kernel_0.shape
        # get the center to not rotate around the (0, 0) coord point
        center = (w // 2, h // 2)
        # apply rotation
        rotation_matrix = cv2.getRotationMatrix2D(center, angle, 1)
        kernel_angle_increased = cv2.warpAffine(
            thin_kernel_0, rotation_matrix, (w, h), cv2.INTER_NEAREST
        )

        # get the k=3 kerne
        kernel_angle = kernel_angle_increased[1:-1, 1:-1]
        is_diag = abs(kernel_angle) == 1  # because of the interpolation
        kernel_angle = kernel_angle * is_diag  # because of the interpolation
        thin_kernels.append(kernel_angle)
    return thin_kernels



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231013

import os
from random import randint
from typing import Optional

import click
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from neetbox._protocol import VERSION
from neetbox.client._client_web_apis import *
from neetbox.config.project import (
    _get_module_level_config,
    _init_workspace,
    _load_workspace_config,
)
from neetbox.config.user import get as get_global_config
from neetbox.config.user import set as set_global_config
from neetbox.logging import Logger
from neetbox.utils.localstorage import get_create_neetbox_config_directory
from neetbox.utils.massive import check_read_toml

console = Console()

logger = Logger("NEETBOX CLI", skip_writers_names=["ws", "file"])


def get_client_config():
    return _get_module_level_config("client")


# def get_base_addr(port=0):
#     daemon_config = get_daemon_config()
#     _port = port or daemon_config["port"]
#     daemon_address = f"{daemon_config['host']}:{_port}"
#     return f"http://{daemon_address}/"


@click.group()
@click.option(
    "--verbose",
    "-v",
    help="set verbose flag to True",
    default=False,
)
@click.pass_context
def main(ctx, verbose: bool):
    ctx.ensure_object(dict)
    """This is NEETBOX cli"""
    ctx.obj["verbose"] = verbose


def _try_load_workspace_if_applicable():
    CONFIG_FILE_NAME = f"neetbox.toml"
    is_workspace = check_read_toml(CONFIG_FILE_NAME)
    if is_workspace:
        _load_workspace_config(load_only=True)


@main.command(name="version")
def version_command():
    print(VERSION)


@main.command(name="serve")
@click.option(
    "--port", "-p", help="specify which port to launch", metavar="port", required=False, default=0
)
@click.option("--debug", "-d", is_flag=True, help="Run with debug mode", default=False)
def serve(port, debug):
    """serve neetbox server in attached mode"""
    _try_load_workspace_if_applicable()
    _daemon_config = get_client_config()
    try:
        if port:
            _daemon_config["port"] = port
        logger.log(f"Launching server using config: {_daemon_config}")
        from neetbox.server.server_process import server_process

        server_process(cfg=_daemon_config, debug=debug)
    except Exception as e:
        logger.err(f"Failed to launch a neetbox server: {e}", reraise=True)


@main.command(name="shutdown")
@click.option(
    "--port",
    "-p",
    help="specify which port to send shutdown",
    metavar="port",
    required=False,
    default=0,
)
def shutdown_server(port):
    """shutdown neetbox server on specific port"""
    _try_load_workspace_if_applicable()
    try:
        daemon_config = get_client_config()
        if port:
            daemon_config["port"] = port
        _response = shutdown()
        click.echo(_response)
    except Exception as e:
        logger.err(
            f"Failed to shutdown the neetbox server: {e}. Is the server running on port {daemon_config['port']}?"
        )


def console_banner(text, font: Optional[str] = None):
    from pyfiglet import Figlet, FigletFont

    builtin_font_list = [
        "ansiregular",
        "ansishadow",
        "isometrixc2",
        "nscripts",
        "nvscript",
    ]
    if not font:
        font = builtin_font_list[randint(0, len(builtin_font_list)) - 1]

    if font not in FigletFont.getFonts():
        if font in builtin_font_list:  # builtin but not installed
            module_path = os.path.dirname(__file__)
            FigletFont.installFonts(f"{module_path}/flfs/{font}.flf")
        else:  # path?
            assert os.path.isfile(font), "The provided font is not a fontname or a font file path."
            file_name = os.path.basename(font)
            file = os.path.splitext(file_name)
            if file[0] not in FigletFont.getFonts():  # no installed file match the file name
                try:
                    FigletFont.installFonts(f"res/flfs/{font}.flf")
                except Exception:
                    font = None
            else:
                font = file[0]
    f = Figlet(font)
    rendered_text = f.renderText(text)
    console.print(Panel.fit(f"{rendered_text}", border_style="green"))


@main.command(name="init")
@click.option("--name", "-n", help="set project name", metavar="name", required=False)
def init(name: str):
    """initialize current folder as workspace and generate the config file from defaults"""
    try:
        init_succeed = _init_workspace(name=name)
        if init_succeed:
            console_banner("neetbox", font="ansishadow")
            logger.log("Welcome to NEETBOX")
    except Exception as e:
        logger.err(f"Failed to init here: {e}")


@main.command(name="config")
@click.option(
    "--set",
    "kvs",
    type=(str, str),
    multiple=True,
    help="specify which config name to modify as well as target value",
    metavar="name value",
    required=False,
    default=None,
)
@click.option("--force", "-f", is_flag=True, help="force set", default=False)
def configs(kvs, force):
    """list or set global config"""
    if kvs:
        current_configs = get_global_config()
        for k, v in kvs:
            if k not in current_configs and not force:
                logger.warn(
                    f"config key {k} not exist in current config file and will not be writen into config. if you want to write anyway, try --force."
                )
            else:
                set_global_config(k, v)
    table = Table(title="current global configs")
    table.add_column("name", style="green", no_wrap=True)
    table.add_column("current value")
    for k, v in get_global_config().items():
        table.add_row(str(k), str(v))
    console.print(table)
    logger.info(f"from config file folder: {get_create_neetbox_config_directory()}")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240109

from typing import Optional, Union

from fastapi import APIRouter, Body, File, Form, HTTPException, Response, UploadFile

from neetbox._protocol import *
from neetbox.logging import Logger, LogLevel

from ....db.project.condition import ProjectDbQueryCondition
from ._bridge import Bridge

logger = Logger("Project APIs", skip_writers_names=["ws"])
logger.log_level = LogLevel.DEBUG

router = APIRouter()


@router.get(f"/list")
async def get_status_of_all_proejcts():
    return [_project_status_from_bridge(bridge) for _, bridge in Bridge.items()]


@router.get(f"/{{project_id}}")
async def get_status_of_project_id(project_id: str):
    if not Bridge.has(project_id):
        raise HTTPException(status_code=404, detail={ERROR_KEY: "Project ID not found"})
    bridge = Bridge.of_id(project_id)
    return _project_status_from_bridge(bridge)


def _project_status_from_bridge(bridge: Bridge):
    run_id_info_list = bridge.get_run_ids()
    name_of_project = None
    for run_id_info in reversed(run_id_info_list):
        config = bridge.get_status(run_id=run_id_info[RUN_ID_KEY], series="config")
        if NAME_KEY in config:
            name_of_project = config[NAME_KEY]
            break
    return {
        PROJECT_ID_KEY: bridge.project_id,
        "storage": bridge.historyDB.size,
        "online": bridge.is_online(),
        NAME_KEY: name_of_project,
        "runids": run_id_info_list,
    }


def get_history_json_of(project_id: str, table_name: str, condition=Union[dict, str]):
    if not Bridge.has(project_id):
        raise HTTPException(status_code=404, detail={ERROR_KEY: "Project ID not found"})
    try:
        if isinstance(condition, str):
            condition = json.loads(condition)
        if isinstance(condition, dict):
            condition = ProjectDbQueryCondition.loads(condition)
    except Exception as e:  # if failed to parse
        error_message = f"failed to parse condition from {type(condition)}{condition} :{e}"
        logger.debug(error_message, series="400")
        raise HTTPException(status_code=400, detail={ERROR_KEY: error_message})
    return Bridge.of_id(project_id).read_json_from_history(
        table_name=table_name, condition=condition
    )


@router.get(f"/{{project_id}}/log")
async def get_history_log_of(project_id: str, condition: str):
    return get_history_json_of(
        project_id=project_id,
        table_name=LOG_TABLE_NAME,
        condition=condition,
    )


@router.get(f"/{{project_id}}/hardware")
async def get_history_hardware_info_of(project_id: str, condition: str):
    return get_history_json_of(
        project_id=project_id,
        table_name=EVENT_TYPE_NAME_HARDWARE,
        condition=condition,
    )


@router.get(f"/{{project_id}}/series/{{table_name}}")
async def get_series_list_of(
    project_id: str, table_name: str, run_id: str = None
):  # client side function
    if not Bridge.has(project_id):
        raise HTTPException(status_code=404, detail={ERROR_KEY: "Project ID not found"})
    result = Bridge.of_id(project_id).get_series_of(table_name, run_id=run_id)
    return result


@router.get(f"/{{project_id}}/run/{{run_id}}")
async def get_status_of(project_id, run_id):
    if not Bridge.has(project_id):
        raise HTTPException(status_code=404, detail={ERROR_KEY: "Project ID not found"})
    result = Bridge.of_id(project_id).get_status(run_id=run_id)
    return result


@router.put(f"/{{project_id}}/run/{{run_id}}")
async def set_get_metadata_of_run_id(project_id: str, run_id: str, metadata: dict = Body(...)):
    bridge = Bridge.of_id(project_id)
    try:
        old_metadata = bridge.historyDB.fetch_metadata_of_run_id(run_id=run_id)  # get old metadata
        old_metadata.update(metadata)
        # Assuming the method to update metadata in your database might look like this
        bridge.historyDB.update_metadata_of_run_id(run_id=run_id, metadata=old_metadata)
        return bridge.historyDB.fetch_metadata_of_run_id(run_id=run_id)
    except Exception as e:  # Replace with your specific database exception
        raise HTTPException(status_code=404, detail={ERROR_KEY: str(e)})


@router.delete(f"/{{project_id}}/run/{{run_id}}")
async def delete_run_id(project_id: str, run_id: str):
    bridge = Bridge.of_id(project_id)
    if bridge.is_online(run_id):  # cannot delete running projects
        raise HTTPException(status_code=400, detail={ERROR_KEY: "can only delete history run id."})
    bridge.historyDB.delete_run_id(run_id)
    if 0 == len(bridge.get_run_ids()):  # check if all the run ids are deleted
        del Bridge._id2bridge[project_id]  # delete the empty bridge
    return {RESULT_KEY: "success"}


@router.post(f"/{{project_id}}/image")
async def upload_image(project_id: str, image: UploadFile = File(...), metadata: str = Form(...)):
    if not Bridge.has(project_id):
        raise HTTPException(status_code=404, detail={ERROR_KEY: "project id not found"})
    message = EventMsg.loads(metadata)
    image_bytes = await image.read()
    message.id = Bridge.of_id(project_id).save_blob_to_history(
        table_name="image",
        run_id=message.run_id,
        series=message.series,
        meta_data=message.payload,
        timestamp=message.timestamp,
        blob_data=image_bytes,
        num_row_limit=message.history_len,
    )
    message.payload = message.payload or {}
    await Bridge.of_id(project_id).ws_send_to_frontends(message)
    return {RESULT_KEY: "ok", ID_KEY: message.id}


@router.get(f"/{{project_id}}/image/{{image_id}}")
async def get_image_of(project_id: str, image_id: int, meta: Optional[bool] = None):
    if not Bridge.has(project_id):
        raise HTTPException(status_code=404, detail={ERROR_KEY: "project id not found"})
    # Database logic here
    [(_, _, meta_data, image)] = Bridge.of_id(project_id).read_blob_from_history(
        table_name="image", condition=ProjectDbQueryCondition(id=image_id), meta_only=meta
    )
    if meta:
        return Response(meta_data, media_type="application/json")
    else:
        return Response(image, media_type="image/png")


@router.get(f"/{{project_id}}/image")
async def get_history_image_metadata_of(project_id: str, condition: Optional[str] = None):
    if not Bridge.has(project_id):
        raise HTTPException(status_code=404, detail={ERROR_KEY: "project id not found"})
    try:
        condition_json = json.loads(condition) if condition else {}
        condition = ProjectDbQueryCondition.loads(condition_json)
    except Exception as e:
        raise HTTPException(status_code=400, detail={ERROR_KEY: str(e)})
    query_results = Bridge.of_id(project_id).read_blob_from_history(
        table_name="image", condition=condition, meta_only=True
    )
    result = [{"imageId": id, "metadata": meta_data} for (id, _, meta_data) in query_results]
    return result


@router.get(f"/{{project_id}}/scalar")
async def get_history_scalar_of(project_id: str, condition: str = None):
    try:
        condition_json = json.loads(condition) if condition else "{}"
        condition = ProjectDbQueryCondition.loads(condition_json)
    except Exception as e:
        raise HTTPException(status_code=400, detail={ERROR_KEY: str(e)})
    return get_history_json_of(project_id=project_id, table_name="scalar", condition=condition)


@router.get(f"/{{project_id}}/progress")
async def get_history_progress_of(project_id: str, condition: str = None):
    try:
        condition_json = json.loads(condition) if condition else "{}"
        condition = ProjectDbQueryCondition.loads(condition_json)
    except Exception as e:
        raise HTTPException(status_code=400, detail={ERROR_KEY: str(e)})
    return get_history_json_of(project_id=project_id, table_name="progress", condition=condition)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240120

import os
import socket
import time
from threading import Thread

from fastapi import APIRouter

from neetbox._protocol import *
from neetbox.logging import Logger, LogLevel

logger = Logger("Project APIs", skip_writers_names=["ws"])
logger.log_level = LogLevel.DEBUG

router = APIRouter()


@router.get("/hello")
async def just_send_hello():
    return {"hello": "hello"}


@router.get("/listips")
async def list_server_ips():
    local_hostname = socket.gethostname()
    ip_addresses = socket.gethostbyname_ex(local_hostname)[2]
    filtered_ips = [ip for ip in ip_addresses if not ip.startswith("127.")]
    return {"hostname": local_hostname, "ips": filtered_ips}


@router.post(f"/shutdown")
async def shutdown_server():
    def __sleep_and_shutdown(secs=1):
        time.sleep(secs)
        os._exit(0)

    Thread(target=__sleep_and_shutdown).start()  # shutdown after 3 seconds
    logger.info(f"BYE.")
    return {RESULT_KEY: f"shutdown in 1 seconds."}



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231216

import collections
import functools
import os
import socket
import struct
from concurrent.futures import ThreadPoolExecutor

_ThreadPoolExecutor = ThreadPoolExecutor()


def is_loopback(host):
    loopback_checker = {
        socket.AF_INET: lambda x: struct.unpack("!I", socket.inet_aton(x))[0] >> (32 - 8) == 127,
        socket.AF_INET6: lambda x: x == "::1",
    }
    for family in (socket.AF_INET, socket.AF_INET6):
        try:
            r = socket.getaddrinfo(host, None, family, socket.SOCK_STREAM)
        except socket.gaierror:
            return False
        for family, _, _, _, sockaddr in r:
            if not loopback_checker[family](sockaddr[0]):
                return False
    return True


def nonblocking(func):
    @functools.wraps(func)
    def _use_thread_pool(*args, **kwargs):
        try:
            return _ThreadPoolExecutor.submit(func, *args, **kwargs)  # type: ignore
        except Exception as e:
            raise e

    return _use_thread_pool


def update_dict_recursively(self: dict, the_other: dict):
    for _k, _v in the_other.items():
        if type(_v) is dict:  # currently resolving a dict child
            if _k in self:  # dfs merge
                update_dict_recursively(self=self[_k], the_other=the_other[_k])
            else:
                self[_k] = the_other[_k]
        else:  # not a dict, overwriting
            self[_k] = the_other[_k]


def check_read_toml(path) -> bool:
    import toml

    if not os.path.isfile(path):  # but config file not exist
        return False
    try:
        return toml.load(path)  # try load as toml
    except Exception as e:
        return False


def describe_object(obj, length_limit=None) -> str:
    if hasattr(obj, "__name__"):
        description = obj.__name__
    elif hasattr(obj, "__class__"):
        description = f"{obj.__class__.__name__} object"
    else:
        description = str(obj)
    if length_limit and len(description) > length_limit:
        description = description[:length_limit] + "..."
    return description



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240110

import time

from fastapi import APIRouter, WebSocket, WebSocketDisconnect

from neetbox._protocol import *
from neetbox.logging import Logger

from ._monitor import disks, hardware

router = APIRouter()

logger = Logger("Websocket.Server", skip_writers_names=["ws"])

WS_CLIENTS = []


async def handshake(websocket: WebSocket):
    await websocket.accept()
    message = await websocket.receive_text()  # the first message should be handshake message
    try:
        message = EventMsg.loads(message)
    except Exception as e:
        logger.err(f"error encountered '{e}' while handling handshake message: {message}")
        websocket.close(code=1003, reason="invalid handshake")
        return None
    assert (
        message.event_type == EVENT_TYPE_NAME_HANDSHAKE
    ), f"handshake expected but got {message.event_type}"
    # assign this client to a Bridge
    logger.info(f"handling handshake for {message.identity_type}")
    assert message.identity_type == IdentityType.WEB
    await websocket.send_json(
        data=EventMsg.merge(
            message,
            {
                PAYLOAD_KEY: {RESULT_KEY: 200, REASON_KEY: "join success"},
                IDENTITY_TYPE_KEY: IdentityType.SERVER,
            },
        ).json
    )
    WS_CLIENTS.append(websocket)
    return websocket


@router.websocket("/")
async def server_monitor_ws_endpoint(websocket: WebSocket):
    if not await handshake(websocket):
        return
    try:
        sec_counter = 0
        while True:
            await websocket.send_json(
                EventMsg(
                    project_id=None,
                    run_id=None,
                    event_type=EVENT_TYPE_NAME_HARDWARE,
                    identity_type=IdentityType.SERVER,
                    payload=hardware.json,
                    timestamp=get_timestamp(),
                ).json
            )
            time.sleep(1)
            sec_counter += 1
            if sec_counter >= 9:
                await websocket.send_json(
                    EventMsg(
                        project_id=None,
                        run_id=None,
                        event_type=EVENT_TYPE_NAME_HARDWARE,
                        identity_type=IdentityType.SERVER,
                        payload=disks.json,
                        timestamp=get_timestamp(),
                    ).json
                )
                sec_counter = 0

    except WebSocketDisconnect:
        try:
            WS_CLIENTS.remove(websocket)
        except:
            pass



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230315

import time

import torch
from tqdm import tqdm

from neetbox.logging import logger
from neetbox.utils import pkg


def profile(
    model,
    input_shape=(1, 3, 1280, 720),
    specific_input=None,
    profiling=True,
    speedtest=1000,
):
    assert pkg.is_installed("thop", try_install_if_not=True)
    if speedtest:
        input_tensor = specific_input
        if not input_tensor:
            input_tensor = torch.rand(input_shape)
            if next(model.parameters()).is_cuda:
                input_tensor = input_tensor.cuda()
        counter = []
        logger.log("running speedtest...")
        model.eval()
        with torch.no_grad():
            logger.log("start warm up")
            for i in tqdm(range(10)):
                model(input_tensor)
            logger.log("warm up done")
            for test_index in tqdm(range(speedtest + 2)):
                t = time.time()
                output = model(input_tensor)
                counter.append(time.time() - t)
            _min, _max, _sum = min(counter), max(counter), sum(counter)
        _aver = (_sum - _min - _max) / speedtest
        fps_str = f"average {_aver}s per run"
        if type(output) in [list, torch.Tensor]:
            if type(output) is list:
                output_shape = f"{len(output)}-element-list\n"
            else:
                output_shape = output.shape
            fps_str += f" with out put size {output_shape}"
        logger.log(fps_str)
        logger.log(f"min inference time: {_min}s, Max inference time: {_max}s")
        _fps_aver = 1.0 / _aver
        logger.log(f"That is {_fps_aver} frames per second")

        if next(model.parameters()).is_cuda:
            time_counter = torch.zeros((speedtest + 2, 1))
            time0, time1 = torch.cuda.Event(enable_timing=True), torch.cuda.Event(
                enable_timing=True
            )
            model.eval()
            with torch.no_grad():
                logger.log("running CUDA side synchronous speedtest...")
                for test_index in tqdm(range(speedtest + 2)):
                    time0.record()
                    output = model(input_tensor)
                    time1.record()
                    torch.cuda.synchronize()
                    time_counter[test_index] = time0.elapsed_time(time1)
                _min, _max, _sum = (
                    time_counter.min(),
                    time_counter.max(),
                    time_counter.sum(),
                )
                _aver = (_sum - _min - _max) / 1000.0 / speedtest
                fps_str = f"average {_aver}s per run"
                if type(output) in [list, torch.Tensor]:
                    if type(output) is list:
                        output_shape = f"{len(output)}-element-list\n"
                    else:
                        output_shape = output.shape
                    fps_str += f" with out put size {output_shape}"
                logger.log(fps_str)
                logger.log(
                    f"min inference time: {_min / 1000.}s, max inference time: {_max / 1000.}s"
                )
                logger.log(f"That is {1. / _aver} frames per second")
                if _fps_aver - 1.0 / _aver > _fps_aver * 0.05:
                    logger.warn(
                        f"Seems your model has an imbalanced performance peek between CUDA side synchronous test and none-sync one. Consider raising speedtest loop times (currently {speedtest} +2) to have a stable result."
                    )
                logger.debug(
                    "Note that the CUDA side synchronous speedtest is more reliable since you are using a GPU."
                )
    if profiling:
        logger.log("model profiling...")
        input_tensor = specific_input
        if not input_tensor:
            input_tensor = torch.rand(input_shape)
            if next(model.parameters()).is_cuda:
                input_tensor = input_tensor.cuda()
        input_tensor = (input_tensor,)
        assert pkg.is_installed("thop")
        from thop import profile as _profile

        flops, params = _profile(model, inputs=input_tensor)
        logger.log(f"Model FLOPs = {flops/1e9}G, params = {params/1e6}M")



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240116

import json
from typing import Dict, Tuple, Union

from neetbox._protocol import *

from ..abc import SortType


class ProjectDbQueryCondition:
    def __init__(
        self,
        id: Union[Tuple[int, int], int] = None,
        timestamp: Union[Tuple[str, str], str] = None,
        series: str = None,
        run_id: Union[str, int] = None,
        limit: int = None,
        order: Dict[str, SortType] = {},
    ) -> None:
        self.id_range = id if isinstance(id, tuple) else (id, None)
        self.timestamp_range = timestamp if isinstance(timestamp, tuple) else (timestamp, None)
        self.series = series
        self.run_id = run_id
        self.limit = limit
        self.order = {order[0], order[1]} if isinstance(order, tuple) else order

    @classmethod
    def loads(cls, json_data):
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        """
        {
            "id" : [int,int], # from,to
            "timestamp" : [str,str], # from,to
            "series" : str, # series name
            "limit" : int,
            "order" : [
                {"column name" : "ASC/DESC"},
                ...
            ]
        }
        """
        # try load id range
        id_range = None
        if ID_COLUMN_NAME in json_data:
            id_range_str = json_data[ID_COLUMN_NAME]
            id_range = eval(id_range_str)
            assert (
                isinstance(id_range, list)
                and len(id_range) == 2
                and type(id_range[0]) is int
                and type(id_range[1]) is int
                or type(id_range) is int
            )
            id_range = tuple(id_range)  # to tuple
        # try load timestamp range
        timestamp_range = None
        if TIMESTAMP_COLUMN_NAME in json_data:
            timestamp_range_str = json_data[TIMESTAMP_COLUMN_NAME]
            timestamp_range = eval(timestamp_range_str)
            assert (
                isinstance(timestamp_range, list)
                and len(timestamp_range) == 2
                and type(timestamp_range[0]) is str
                and type(timestamp_range[1]) is str
                or type(timestamp_range) is str
            )
            # datetime.strptime(timestamp_range[0], DATETIME_FORMAT) # try parse to datetime, makesure its valid
            # datetime.strptime(timestamp_range[1], DATETIME_FORMAT)
            timestamp_range = tuple(timestamp_range)
        # try to load series
        series = json_data[SERIES_COLUMN_NAME] if SERIES_COLUMN_NAME in json_data else None
        # run-id cond
        run_id = json_data[RUN_ID_COLUMN_NAME] if RUN_ID_COLUMN_NAME in json_data else None
        # try load limit
        limit = None
        if "limit" in json_data:
            limit = json_data["limit"]
            assert type(limit) is int
        # try load order
        order = None
        if "order" in json_data:
            order = json_data["order"]
            assert isinstance(order, dict)
        return ProjectDbQueryCondition(
            id=id_range,
            timestamp=timestamp_range,
            series=series,
            run_id=run_id,
            limit=limit,
            order=order,
        )

    def dumpt(self):
        query_cond_vars = []
        # === id condition ===
        _id_cond_str = ""
        if self.id_range[0]:
            _id_0, _id_1 = self.id_range
            if _id_1 is None:
                _id_cond_str = f"{ID_COLUMN_NAME} = ?"
                query_cond_vars.append(_id_0)
            else:
                _id_cond_str = f"{ID_COLUMN_NAME} BETWEEN ? AND ?"
                query_cond_vars.append(_id_0)
                query_cond_vars.append(_id_1)
        # === timestamp condition ===
        _timestamp_cond_str = ""
        if self.timestamp_range[0]:
            _ts_0, _ts_1 = self.timestamp_range
            if _ts_1 is None:
                _timestamp_cond_str = f"{TIMESTAMP_COLUMN_NAME} >= ?"
                query_cond_vars.append(_ts_0)
            else:
                _timestamp_cond_str = f"{TIMESTAMP_COLUMN_NAME} BETWEEN ? AND ?"
                query_cond_vars.append(_ts_0)
                query_cond_vars.append(_ts_1)
        # === series condition ===
        _series_cond_str = ""
        if self.series:
            _series_cond_str = f"{SERIES_COLUMN_NAME} = ?"
            query_cond_vars.append(self.series)
        # === run-id condition ===
        _run_id_cond_str = ""
        if self.run_id:
            _run_id_cond_str = f"{RUN_ID_COLUMN_NAME} = ?"
            query_cond_vars.append(self.run_id)
        # === ORDER BY ===
        _order_cond = f"ORDER BY " if self.order else ""
        if self.order:
            for _col_name, _sort in self.order.items():
                _order_cond += (
                    f"{_col_name} {_sort.value if isinstance(_sort,SortType) else _sort}, "
                )
            _order_cond = _order_cond[:-2]  # remove last ','
        # === LIMIT ===
        _limit_cond_str = f"LIMIT {self.limit}" if self.limit else ""
        # === concat conditions ===
        query_condition_strs = []
        for cond in [_id_cond_str, _timestamp_cond_str, _series_cond_str, _run_id_cond_str]:
            if cond:
                query_condition_strs.append(cond)
        query_condition_strs = " AND ".join(query_condition_strs)
        # === concat order by and limit ===
        order_and_limit = []
        for cond in [_order_cond, _limit_cond_str]:
            if cond:
                order_and_limit.append(cond)
        order_and_limit = " ".join(order_and_limit)
        # result
        if query_condition_strs:
            query_condition_strs = f"WHERE {query_condition_strs}"
        query_cond_str = f"{query_condition_strs} {order_and_limit}"
        return query_cond_str, query_cond_vars



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230315

import os
import pathlib
import signal
import threading
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from random import random
from threading import Event
from typing import Dict, List, Union
from urllib.request import urlopen

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

_loader_pool: Dict[str, "ResourceLoader"] = dict()  # all ResourceLoaders are stored here


class ResourceLoader:
    ready: bool = False  # stands for each scan
    file_path_list: list = []
    _initialized: bool = False  # stands for the first scan on creation

    def __new__(
        cls,
        folder: str = ".",
        file_types=["*"],
        sub_dirs=True,
        verbose=False,
        force_rescan=False,
        *args,
        **kwargs,
    ):
        _id = folder + str(file_types) + "_R" if sub_dirs else ""
        if _id in _loader_pool and not force_rescan:
            # logger.info(
            #     "ResourceLoader with same path and same file type(s) already exists. Returning the old one."
            # )
            pass
        else:
            _loader_pool[_id] = super(ResourceLoader, cls).__new__(cls)
        return _loader_pool[_id]

    def __init__(
        self,
        folder: str = ".",
        file_types=["*"],
        sub_dirs=True,
        async_scan=False,
        verbose=False,
        force_rescan=False,
    ):
        """ResourceLoader scans given file type(s) in given place(s)

        Args:
            folder (str): which folder to scan
            file_types (str[]): file type(s) to include. For example, ['jpg','png']
            sub_dirs (bool, optional): scan sub-folder(s)?. Defaults to True.
            async_scan (bool, optional): run scan traks in a new thread. Defaults to False.
            verbose (bool, optional): verbose output. Defaults to False.
            force_rescan (bool, optional): rescan the folder even the same file type(s) was scanned here before. Default to False.
        """
        super().__init__()
        self.path = os.path.abspath(folder)
        self._file_types = file_types
        self._scan_sub_dirs = sub_dirs
        self._async_scan = async_scan
        if not self.ready:
            self._scan(verbose)

    def _scan(self, verbose):
        if not self.ready and self._initialized:
            raise Exception("another scanning requested during the previous one.")
        self.ready = False

        def can_match(path: pathlib.Path):
            if not path.is_file():
                return False
            for file_type in self._file_types:
                if path.match("*." + file_type):
                    return True
            return False

        def perform_scan():
            glob_str = "**/*" if self._scan_sub_dirs else "*"
            if not verbose:  # do not output
                self.file_path_list = [
                    str(path) for path in pathlib.Path(self.path).glob(glob_str) if can_match(path)
                ]
            else:
                self.file_path_list = []
                for path in pathlib.Path(self.path).glob(glob_str):
                    if can_match(path):
                        self.file_path_list.append(path)
            self.ready = True  # scan complete
            if not self._initialized:
                self._initialized = True
            # logger.ok(
            #     f"Resource loader '{self.path}' ready with {'all' if '*' in self._file_types else len(self._file_types)} file types({len(self.file_path_list)} files)."
            # )

        # logger.log(
        #     f"Scanning started at '{self.path}' for {'all' if '*' in self._file_types else len(self._file_types)} file types."
        # )
        # call to scan
        if self._async_scan:
            threading.Thread(target=perform_scan).start()
        else:
            perform_scan()

    def get_file_list(self):
        if not self.ready:
            raise Exception("not ready. scanning in process.")
        return self.file_path_list.copy()

    def __getitem__(self, index):
        if not self.ready:
            raise Exception("not ready. scanning in process.")
        if type(index) is int:
            return self.file_path_list[index]


# class ImagesLoader(ResourceLoader):
#     def __init__(
#         self,
#         folder,
#         file_types=["png", "jpg"],
#         sub_dirs=True,
#         async_scan=False,
#         verbose=False,
#     ):
#         pkg.is_installed("PIL", try_install_if_not="pillow")

#         super().__init__(folder, file_types, sub_dirs, async_scan, verbose)

#     def get_random_image(self):
#         rand_img_path = self.file_path_list[int(random() * len(self.file_path_list))]
#         image = Image.open(rand_img_path).convert("RGB")
#         return image

#     def get_random_images(self, howmany=1):
#         assert howmany < len(self.file_path_list)
#         rand_idx_begin = int(random() * (len(self.file_path_list) - howmany))
#         image_path_list = self[rand_idx_begin : rand_idx_begin + howmany]
#         image_list = []
#         for path in image_path_list:
#             image_list.append(Image.open(path).convert("RGB"))
#         return image_list

#     def get_random_image_as_numpy(self):
#         import numpy as np
#         image = self.get_random_image()
#         return np.array(image)

#     def get_random_image_as_tensor(self, engine="torch"):
#         assert engine in ["torch"]  # TODO support other engines ?
#         assert pkg.is_installed("torchvision")
#         import torchvision.transforms as T

#         tensor_transform = T.Compose(
#             [
#                 T.ToTensor(),
#                 T.Normalize(mean=(0.5, 0.5, 0.5), std=(0.5, 0.5, 0.5)),
#             ]
#         )
#         image = tensor_transform(self.get_random_image()).unsqueeze(0)  # To tensor of NCHW
#         return image


def download(
    urls: Union[str, List[str], Dict[str, str]],
    filenames: Union[str, List[str]] = None,
    max_workers=4,
):
    """download both online and local files from urls

    Example: download('')

    Args:
        urls (Union[str, List[str], Dict[str, str]]): single str to download from url; list of strs to download from urls; dict(filename:url) to download urls and rename them to filenames
        filenames (Union[str, List[str]], optional): str to rename the downloaded file; list of strs to rename downloaded files; None means no rename. Defaults to None.
        overwrite Bool: whether to skip exist files. Default to True.
        retry (int, optional): retries when error occures. Defaults to 3.
    """
    progress = Progress(
        TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
    )

    done_event = Event()

    def handle_sigint(signum, frame):
        done_event.set()

    signal.signal(signal.SIGINT, handle_sigint)

    def copy_url(task_id: TaskID, url: str, path: str) -> None:
        """Copy data from a url to a local file."""
        response = urlopen(url)
        # This will break if the response doesn't contain content length
        progress.update(task_id, total=int(response.info()["Content-length"]))
        with open(path, "wb") as dest_file:
            progress.start_task(task_id)
            for data in iter(partial(response.read, 32768), b""):
                dest_file.write(data)
                progress.update(task_id, advance=len(data))
                if done_event.is_set():
                    return
        # logger.log(f"Downloaded {path}.")

    assert type(urls) in [str, list, dict], "unkown format of url"

    if type(urls) is str:
        assert (
            filenames is None
            or type(filenames) is str
            or (type(filenames) is list and len(filenames) == 1)
        ), "num of urls and filenames mismatch"
        urls = [
            urls,
        ]
        filenames = (
            [
                filenames,
            ]
            if filenames
            else None
        )
    if type(urls) is list:
        if filenames is None:
            filenames = [None] * len(urls)
        assert type(filenames) is list and len(urls) == len(
            filenames
        ), "num of urls and filenames mismatch"

        urls = {filenames[i]: urls[i] for i in range(len(urls))}

    # logger.log(f"Downloading {len(urls)} file(s)...")

    with progress:
        with ThreadPoolExecutor(max_workers=max_workers) as pool:
            for fname, url in urls.items():
                filename = url.split("/")[-1]
                dest_path = os.path.join(fname)
                task_id = progress.add_task("download", filename=filename, start=False)
                pool.submit(copy_url, task_id, url, dest_path)

    return urls



import array
import os
import sys
import zipfile

import torch
from tqdm import tqdm

from neetbox.logging import logger
from neetbox.utils import download

__all__ = ["load_word_vectors"]


def load_word_vectors(root, wv_type, dim):
    """
    Load word vectors from a path, trying .pt, .txt, and .zip extensions.
    Args:
        root: the root path of word vectors.
        wv_type: the type of word vectors, current support [glove.42B, glove.840B, glove.twitter.27B, glove.6B].
        dim: the dimension of word vectors.

    Returns:
        tuple: the tuple of (word vector dict, word vector array, word vector size)
    """
    URL = {
        "glove.42B": "http://nlp.stanford.edu/data/glove.42B.300d.zip",
        "glove.840B": "http://nlp.stanford.edu/data/glove.840B.300d.zip",
        "glove.twitter.27B": "http://nlp.stanford.edu/data/glove.twitter.27B.zip",
        "glove.6B": "http://nlp.stanford.edu/data/glove.6B.zip",
    }
    AVAILABLE = {
        "glove.42B": [300],
        "glove.840B": [300],
        "glove.twitter.27B": [25, 50, 100, 200],
        "glove.6B": [50, 100, 200, 300],
    }

    if wv_type not in AVAILABLE.keys():
        raise ValueError(
            f"no word vector type available, only available for {list(AVAILABLE.keys())}."
        )

    if isinstance(dim, int):
        if dim not in AVAILABLE[wv_type]:
            raise ValueError(f"invalid word vector dimension, only support {AVAILABLE[wv_type]} d.")
        dim = str(dim) + "d"
    fname = os.path.join(root, wv_type + "." + dim)

    if os.path.isfile(fname + ".pt"):
        fname_pt = fname + ".pt"
        logger.log(f"loading word vectors from {fname_pt}.")
        try:
            return torch.load(fname_pt, map_location=torch.device("cpu"))
        except Exception as e:
            logger.err("fail to load the model from {fname_pt}{e}.")
            sys.exit(-1)
    else:
        logger.log(f"File not found: {fname}.pt")
    if not os.path.isfile(fname + ".txt"):
        logger.log(f"File not found: {fname}.txt")
    if os.path.isfile(fname + ".txt"):
        fname_txt = fname + ".txt"
        cm = open(fname_txt, "rb")
        cm = [line for line in cm]
    elif os.path.basename(wv_type) in URL:
        url = URL[wv_type]
        logger.log(f"downloading word vectors from {url}.")
        filename = os.path.basename(fname)
        if not os.path.exists(root):
            os.makedirs(root)
        fname = download(url, fname)
        with zipfile.ZipFile(fname, "r") as zf:
            logger.log(f"extracting word vectors into {root}.")
            zf.extractall(root)
        if not os.path.isfile(fname + ".txt"):
            raise RuntimeError("no word vectors of requested dimension found.")
        return load_word_vectors(root, wv_type, dim)
    else:
        raise RuntimeError("unable to load word vectors.")

    wv_tokens, wv_arr, wv_size = [], array.array("d"), None
    if cm is not None:
        for line in tqdm(range(len(cm)), desc="loading word vectors from {}".format(fname_txt)):
            entries = cm[line].strip().split(b" ")
            word, entries = entries[0], entries[1:]
            if wv_size is None:
                wv_size = len(entries)
            try:
                word = word.decode("utf-8")
            except:
                logger.log(f"non-UTF8 token {repr(word)} was ignored.")
                continue
            wv_arr.extend(float(x) for x in entries)
            wv_tokens.append(word)

    wv_dict = {word: i for i, word in enumerate(wv_tokens)}
    wv_arr = torch.Tensor(wv_arr).view(-1, wv_size)
    ret = (wv_dict, wv_arr, wv_size)
    torch.save(ret, fname + ".pt")
    return ret



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231211

import io
from typing import Optional

import cv2
import numpy as np
from PIL import Image

from neetbox._protocol import *
from neetbox.config import get_project_id, get_run_id
from neetbox.logging import logger
from neetbox.utils.x2numpy import *

from .._client import connection

# ===================== IMAGE things ===================== #


def figure_to_image(figures, close=True):
    """Render matplotlib figure to numpy format.

    Note that this requires the ``matplotlib`` package.

    Args:
        figure (matplotlib.pyplot.figure) or list of figures: figure or a list of figures
        close (bool): Flag to automatically close the figure

    Returns:
        numpy.array: image in [CHW] order
    """
    import numpy as np

    try:
        import matplotlib.backends.backend_agg as plt_backend_agg
        import matplotlib.pyplot as plt
    except ModuleNotFoundError:
        print("please install matplotlib")

    def render_to_rgb(figure):
        canvas = plt_backend_agg.FigureCanvasAgg(figure)
        canvas.draw()
        data = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8)
        w, h = figure.canvas.get_width_height()
        image_hwc = data.reshape([h, w, 4])[:, :, 0:3]
        image_chw = np.moveaxis(image_hwc, source=2, destination=0)
        if close:
            plt.close(figure)
        return image_chw

    if isinstance(figures, list):
        images = [render_to_rgb(figure) for figure in figures]
        return np.stack(images)
    else:
        image = render_to_rgb(figures)
        return image


def make_grid(I, ncols=8):
    # I: N1HW or N3HW
    import numpy as np

    assert isinstance(I, np.ndarray), "plugin error, should pass numpy array here"
    if I.shape[1] == 1:
        I = np.concatenate([I, I, I], 1)
    assert I.ndim == 4 and I.shape[1] == 3 or I.shape[1] == 4
    nimg = I.shape[0]
    H = I.shape[2]
    W = I.shape[3]
    ncols = min(nimg, ncols)
    nrows = int(np.ceil(float(nimg) / ncols))
    canvas = np.zeros((I.shape[1], H * nrows, W * ncols), dtype=I.dtype)
    i = 0
    for y in range(nrows):
        for x in range(ncols):
            if i >= nimg:
                break
            canvas[:, y * H : (y + 1) * H, x * W : (x + 1) * W] = I[i]
            i = i + 1
    return canvas


def convert_to_HWC(tensor, input_format):  # tensor: numpy array
    assert len(set(input_format)) == len(
        input_format
    ), "You an not use the same dimension shorthand twice. input_format: {}".format(input_format)
    assert len(tensor.shape) == len(
        input_format
    ), "size of input tensor and input format are different. tensor shape: {}, input_format: {}".format(
        tensor.shape, input_format
    )
    input_format = input_format.upper()

    if len(input_format) == 2:
        index = [input_format.find(c) for c in "HW"]
        tensor = tensor.transpose(index)
        tensor = np.stack([tensor, tensor, tensor], 2)
        return tensor

    if len(input_format) == 3:
        if "N" not in input_format:
            index = [input_format.find(c) for c in "HWC"]
            tensor_HWC = tensor.transpose(index)
            if tensor_HWC.shape[2] == 1:
                tensor_HWC = np.concatenate([tensor_HWC, tensor_HWC, tensor_HWC], 2)
            return tensor_HWC
        else:
            index = [input_format.find(c) for c in "NHW"]
            tensor = tensor.transpose(index)[:, None, :, :]
            input_format = "NCHW"

    if len(input_format) == 4:
        index = [input_format.find(c) for c in "NCHW"]
        tensor_NCHW = tensor.transpose(index)
        tensor_CHW = make_grid(tensor_NCHW)
        return tensor_CHW.transpose(1, 2, 0)


def add_image(name: str, image, dataformats: str = None):
    """send an image to frontend display

    Args:
        image (Union[np.array, Image.Image, Tensor]): image from cv2 and PIL.Image as well as tensors are supported
        name (str): name of the image, used in frontend display
        dataformats (str): if you are passing a tensor as image, please indicate how to understand the tensor. For example, dataformats="NCWH" means the first axis of the tensor is Number of batches, the second axis is Channel, and the third axis is Width, and the fourth axis is Height.

    """
    if isinstance(image, Image.Image):  # is PIL Image
        with io.BytesIO() as image_bytes_stream:
            # convert PIL Image to bytes
            image.save(image_bytes_stream, format="PNG")
            image_bytes = image_bytes_stream.getvalue()
    else:  # try convert numpy
        dataformats = dataformats or "CHW"
        image = make_np(image)
        image = convert_to_HWC(image, dataformats)
        if image.dtype != np.uint8:
            image = (image * 255.0).astype(np.uint8)

        if isinstance(image, np.ndarray):  # convert ndarray to bytes
            _, im_buf_arr = cv2.imencode(".png", image)
            image_bytes = im_buf_arr.tobytes()

    # send bytes
    project_id = get_project_id()
    run_id = get_run_id()
    try:
        message = EventMsg(
            project_id=project_id,
            run_id=run_id,
            identity_type=IdentityType.CLI,
            series=name,
            event_type=EVENT_TYPE_NAME_IMAGE,
        )
        result = connection.post_check_online(
            api=f"{API_ROOT}/project/{project_id}/image",
            data={METADATA_KEY: message.dumps()},
            files={"image": image_bytes},
        )
        response_dict = result.json()
        assert (
            RESULT_KEY in response_dict and response_dict[RESULT_KEY] == "ok"
        ), "server response not ok"
    except Exception as e:
        logger.warn(f"unable to upload image: {e}")


# ===================== MATPLOTLIB things ===================== #


def add_figure(name: str, figure, close: Optional[bool] = True):
    """Render matplotlib figure into an image and add it to summary.
    Note that this requires the ``matplotlib`` package.

    Args:
        tag: Data identifier
        figure (matplotlib.pyplot.figure) or list of figures: Figure or a list of figures
        global_step: Global step value to record
        close: Flag to automatically close the figure
        walltime: Override default walltime (time.time()) of event
    """
    if isinstance(figure, list):
        add_image(name=name, image=figure_to_image(figure, close), dataformats="NCHW")
    else:
        add_image(name=name, image=figure_to_image(figure, close), dataformats="CHW")



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240110

from fastapi import APIRouter, WebSocket, WebSocketDisconnect

from neetbox._protocol import *
from neetbox.logging import Logger

from ._manager import manager

router = APIRouter()
logger = Logger("Websocket.Project", skip_writers_names=["ws"])


@router.websocket("/")
async def project_ws_endpoint(websocket: WebSocket):
    ws_client = await manager.handshake(websocket)
    if not ws_client:
        return
    try:
        while True:
            message = await websocket.receive_text()
            try:  # parse event message
                message = EventMsg.loads(message)
            except Exception as e:
                logger.err(
                    f"Illegal message format from client    {ws_client.id}: {message}, failed to parse cause {e}, dropping..."
                )
                continue
            await manager.handle_event_msg(websocket, message)

    except WebSocketDisconnect:
        manager.disconnect(websocket)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231216

from .client import (
    action,
    add_figure,
    add_hyperparams,
    add_image,
    add_scalar,
    listen,
    watch,
    progress,
)
from .logging import logger
from ._protocol import VERSION as __version__

__all__ = [
    "add_image",
    "add_figure",
    "add_scalar",
    "add_hyperparams",
    "action",
    "logger",
    "watch",
    "listen",
    "progress",
]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230319

import json
import platform
import re


def is_pure_ansi(text: str) -> bool:
    if not re.compile(r"^[A-Za-z0-9_-]+$").match(text):
        return False
    return True


def is_fs_case_sensitive():
    """Check if the file system is case sensitive

    Returns:
        bool: True if case sensitive
    """
    return "windows" not in platform.system().lower()


def legal_file_name_of(text: str) -> str:
    """Remove invalid characters for windows file systems

    Args:
        title (str): the given title

    Returns:
        str: valid text
    """
    if platform.system().lower() == "windows":
        rstr = r"[\/\\\:\*\?\"\<\>\|]"  # '/ \ : * ? " < > |'
        new_title = re.sub(rstr, "_", text)  # replace with '_'
        return new_title
    return text


def is_jsonable(x):
    try:
        x = json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231022

import functools
import json
import logging
import subprocess
import time
from collections import defaultdict
from threading import Lock, Thread
from typing import Callable

import httpx
import websocket

from neetbox._protocol import *
from neetbox.config import get_module_level_config, get_project_id, get_run_id
from neetbox.logging import Logger, RawLog
from neetbox.utils import DaemonableProcess, Registry
from neetbox.utils.massive import is_loopback
from neetbox.utils.mvc import Singleton

logging.getLogger("httpx").setLevel(logging.ERROR)
logger = Logger(name_alias="CLIENT", skip_writers_names=["ws"])


def addr_of_api(api, http_root=None):
    if not http_root:
        config = get_module_level_config()
        daemon_server_address = f"{config['host']}:{config['port']}"
        http_root = f"http://{daemon_server_address}"
    if not api.startswith("/"):
        api = f"/{api}"
    result = f"{http_root}{api}"
    return result


def online_only(func):
    """function decorator which checks client online status on each function call

    Args:
        func (Callable): the function

    Returns:
        Callable: the decorated function
    """

    @functools.wraps(func)
    def __online_only_func(*args, **kwargs):
        if not NeetboxClient()._is_initialized:
            with NeetboxClient()._thread_safe_lock:
                NeetboxClient()._connect_blocking()

        if NeetboxClient()._is_initialized and NeetboxClient().online_mode == False:
            return None
        return func(*args, **kwargs)

    return __online_only_func


class NeetboxClient(metaclass=Singleton):  # singleton
    online_mode: bool = None
    httpxClient: httpx.Client = httpx.Client(  # httpx client
        proxies={
            "http://": None,
            "https://": None,
        }
    )  # type: ignore
    wsApp: websocket.WebSocketApp = None  # websocket client app
    _is_initialized: bool = False
    _thread_safe_lock = Lock()
    is_ws_connected: bool = False
    ws_message_query = []  # websocket message query
    ws_subscribers = defaultdict(list)  # default to no subscribers

    @online_only
    def post_check_online(self, api: str, root: str = None, *args, **kwargs):
        url = addr_of_api(api, http_root=root)
        return self.httpxClient.post(url, *args, **kwargs)

    def post(self, api: str, root: str = None, *args, **kwargs):
        url = addr_of_api(api, http_root=root)
        return self.httpxClient.post(url, *args, **kwargs)

    @online_only
    def get_check_online(self, api: str, root: str = None, *args, **kwargs):
        url = addr_of_api(api, http_root=root)
        return self.httpxClient.get(url, *args, **kwargs)

    def get(self, api: str, root: str = None, *args, **kwargs):
        url = addr_of_api(api, http_root=root)
        return self.httpxClient.get(url, *args, **kwargs)

    @online_only
    def put_check_online(self, api: str, root: str = None, *args, **kwargs):
        url = addr_of_api(api, http_root=root)
        return self.httpxClient.put(url, *args, **kwargs)

    @online_only
    def delete_check_online(self, api: str, root: str = None, *args, **kwargs):
        url = addr_of_api(api, http_root=root)
        return self.httpxClient.delete(url, *args, **kwargs)

    def ws_subscribe(self, event_type_name: str):
        """let a function subscribe to ws messages with event type name.
        !!! dfor inner APIs only, do not use this in your code!
        !!! developers should contorl blocking on their own functions

        Args:
            function (Callable): who is subscribing the event type
            event_type_name (str, optional): Which event to listen. Defaults to None.
        """

        def _ws_subscribe(function: Callable):
            self.ws_subscribers[event_type_name].append(function)
            # logger.info(f"ws: {name} subscribed to '{event_type_name}'")
            return function

        return _ws_subscribe

    def check_server_connectivity(self, config=None):
        config = config or get_module_level_config()
        logger.debug(f"Connecting to daemon at {config['host']}:{config   ['port']} ...")
        daemon_server_address = f"{config['host']}:{config['port']}"
        http_root = f"http://{daemon_server_address}"

        # check if daemon is alive
        def fetch_hello(root):
            response = None
            try:
                response = self.get(api=f"{API_ROOT}/{SERVER_KEY}/hello", root=root)
                print(f"{API_ROOT}{SERIES_KEY}/hello")
                assert response.json()["hello"] == "hello"
            except:
                raise IOError(
                    f"Daemon at {root} is not alive: {response.status_code if response else 'no response'}"
                )

        try:
            fetch_hello(http_root)
            logger.ok(f"daemon alive at {http_root}")
            return True
        except Exception as e:
            logger.err(e)
            return False

    def _connect_blocking(self, config=None):
        if self._is_initialized:
            return  # if already initialized, do nothing

        config = config or get_module_level_config()
        if not config["enable"]:  # check if enable
            self.online_mode = False
            self._is_initialized = True
            return

        if not config["allowIpython"]:  # check if allow ipython
            try:
                eval("__IPYTHON__")  # check if in ipython
            except NameError:  # not in ipython
                pass
            else:  # in ipython
                logger.info(
                    "NEETBOX DAEMON won't start when debugging in ipython console. If you want to allow daemon run in "
                    "ipython, try to set 'allowIpython' to True."
                )
                self.online_mode = False
                self._is_initialized = True
                return  # ignore if debugging in ipython

        server_host = config["host"]
        server_port = config["port"]
        if not self.check_server_connectivity():  # if daemon not online
            if not (
                is_loopback(server_host) or server_host in ["127.0.0.1"]
            ):  # daemon not running on localhost
                logger.err(
                    RuntimeError(
                        f"No daemon running at {server_host}:{server_port}, daemon will not be attached, stopping..."
                    ),
                    reraise=True,
                )
            # connecting localhost but no server alive, create one
            logger.log(
                f"No daemon running on {server_host}:{server_port}, trying to create daemon..."
            )
            import neetbox.server._daemon_server_launch_script as server_launcher

            popen = DaemonableProcess(  # server daemon
                target=server_launcher,
                args=["--config", json.dumps(config)],
                mode=config["mode"],
                redirect_stdout=subprocess.DEVNULL if config["mute"] else None,
                env_append={"NEETBOX_DAEMON_PROCESS": "1"},
            ).start()
            time.sleep(1)
            _retry_timeout = 10
            _time_begin = time.perf_counter()
            logger.debug("Created daemon process, trying to connect to daemon...")
            online_flag = False
            while time.perf_counter() - _time_begin < 10:  # try connect daemon
                if not self.check_server_connectivity():
                    exit_code = popen.poll()
                    if exit_code is not None:
                        logger.err(
                            f"Daemon process exited unexpectedly with exit code {exit_code}."
                        )
                        return False
                    time.sleep(0.5)
                else:
                    online_flag = True
                    break
            if not online_flag:
                logger.err(
                    RuntimeError(
                        f"Failed to connect to daemon after {_retry_timeout}s, stopping..."
                    ),
                    reraise=True,
                )

        self.online_mode = True  # enable online mode
        self.ws_server_url = f"ws://{server_host}:{server_port}{WS_ROOT}/project/"  # ws server url
        logger.info(f"websocket connecting to {self.ws_server_url}")
        self.wsApp = websocket.WebSocketApp(  # create websocket client
            url=self.ws_server_url,
            on_open=self.on_ws_open,
            on_message=self.on_ws_message,
            on_error=self.on_ws_err,
            on_close=self.on_ws_close,
        )

        Thread(
            target=self.wsApp.run_forever, kwargs={"reconnect": 1}, daemon=True
        ).start()  # initialize and start ws thread

        self._is_initialized = True

    def on_ws_open(self, ws: websocket.WebSocketApp):
        project_id = get_project_id()
        logger.ok(f"client websocket connected. sending handshake as '{project_id}'...")
        handshake_msg = EventMsg(  # handshake request message
            project_id=project_id,
            run_id=get_run_id(),
            event_type=EVENT_TYPE_NAME_HANDSHAKE,
            identity_type=IdentityType.CLI,
            event_id=0,
        ).dumps()
        ws.send(handshake_msg)

    def on_ws_err(self, ws: websocket.WebSocketApp, msg):
        logger.err(f"client websocket encountered {msg}")

    def on_ws_close(self, ws: websocket.WebSocketApp, close_status_code, close_msg):
        logger.warn(
            f"client websocket closed: status code: {close_status_code}, message: {close_msg}"
        )
        self.is_ws_connected = False

    def on_ws_message(self, ws: websocket.WebSocketApp, message):
        message = EventMsg.loads(message)  # message should be json
        if message.event_type == EVENT_TYPE_NAME_HANDSHAKE:
            assert message.payload["result"] == 200
            logger.ok(f"neetbox handshake succeed.")
            ws.send(  # send immediately without querying
                EventMsg(
                    project_id=get_project_id(),
                    event_id=message.event_id,
                    event_type=EVENT_TYPE_NAME_STATUS,
                    series="config",
                    run_id=get_run_id(),
                    payload=get_module_level_config("@"),
                ).dumps()
            )
            self.is_ws_connected = True
            # return # DO NOT return!
        if message.event_type not in self.ws_subscribers:
            logger.warn(
                f"Client received a(n) {message.event_type} event but nobody subscribes it. Ignoring anyway."
            )
        for subscriber in self.ws_subscribers[message.event_type]:
            try:
                subscriber(message)  # pass payload message into subscriber
            except Exception as e:
                # subscriber throws error
                logger.err(
                    f"Subscriber {subscriber.__name__} crashed on message event {message.event_type}, ignoring."
                )

    @online_only
    def ws_send(
        self,
        event_type: str,
        payload: dict,
        series=None,
        timestamp: str = None,
        event_id=-1,
        identity_type=IdentityType.CLI,
        _history_len=-1,
    ):
        message = EventMsg(
            project_id=get_project_id(),
            run_id=get_run_id(),
            event_type=event_type,
            event_id=event_id,
            identity_type=identity_type,
            series=series,
            payload=payload,
            timestamp=timestamp or get_timestamp(),
            history_len=_history_len,
        )
        self.ws_message_query.append(message)
        if self.is_ws_connected:  # if ws client exist
            try:
                while len(self.ws_message_query):
                    self.wsApp.send(self.ws_message_query[0].dumps())
                    self.ws_message_query.pop(0)
                return
            except Exception as e:
                pass


# singleton
connection = NeetboxClient()


# assign this connection to websocket log writer
LogWriters = Registry("LOG_WRITERS")


@LogWriters.register(name="ws")
def log_writer_ws(log: RawLog):
    whom = log.caller_name_alias or log.caller_info.format(r"%m > %c > %f")
    payload = {CALLER_ID_KEY: whom, MESSAGE_KEY: log.message}
    connection.ws_send(
        event_type=EVENT_TYPE_NAME_LOG,
        series=log.series,
        payload=payload,
        timestamp=log.timestamp.strftime(r"%Y-%m-%dT%H:%M:%S.%f"),
    )


def _clean_websocket_on_exit():
    # clean websocket connection
    if connection.wsApp is not None:
        connection.wsApp.close()


import atexit

atexit.register(_clean_websocket_on_exit)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230417

import importlib
import pkgutil

from neetbox.config.project import _get_module_level_config as get_module_level_config
from neetbox.config.project import on_config_loaded
from neetbox.utils import Registry
from neetbox.utils.framing import get_frame_module_traceback

__QUERY_AFTER_LOAD_WORKSPACE = Registry("__QUERY_AFTER_LOAD_WORKSPACE")
on_workspace_loaded = __QUERY_AFTER_LOAD_WORKSPACE.register


def _scan_sub_modules():
    __THIS_MODULE = get_frame_module_traceback(1)
    for sub_module_info in pkgutil.iter_modules(__THIS_MODULE.__path__):
        importlib.import_module(f"{__THIS_MODULE.__name__}.{sub_module_info.name}")


@on_config_loaded
def _init_extensions():
    """
    DO NOT call before workspace config load
    """
    if not get_module_level_config()["autoload"]:
        return
    for name, fun in __QUERY_AFTER_LOAD_WORKSPACE.items():
        fun()



import asyncio
import functools
import json
from collections import defaultdict
from dataclasses import dataclass
from typing import Callable, Dict

from rich.console import Console

from .framing import TracebackInfo, get_caller_info_traceback
from .mvc import Singleton

console = Console()
MessageType = str
NameType = str


@dataclass
class MessageListener:
    creator: TracebackInfo
    message_type: str
    name: str
    func: Callable

    @property
    def json(self):
        return {
            "creator": self.creator.json,
            "messageType": self.message_type,
            "listenerName": self.name,
            "function": self.func.__name__,
        }


class Messaging(metaclass=Singleton):
    _listener_dicts: Dict[MessageType, Dict[NameType, MessageListener]] = defaultdict(dict)

    def listener(self, message_type: str, name: str = None):
        caller = get_caller_info_traceback(stack_offset=2)

        def _listener(func: Callable, message_type: str, name: str):
            if asyncio.iscoroutinefunction(func):
                raise RuntimeError(
                    f"cannot add listener func {func}, messaging does not support async listeners."
                )
            name = name or f"{caller.strid}:{func.__name__}"
            if name in self._listener_dicts[message_type]:
                raise RuntimeError(
                    f"message listener for message type {message_type} with name {name} already exist"
                )
            self._listener_dicts[message_type][name] = MessageListener(
                creator=caller, message_type=message_type, name=name, func=func
            )
            return func

        return functools.partial(_listener, message_type=message_type, name=name)

    def send(self, message_type: str, message: any = None):
        if message_type in self._listener_dicts:
            listener: MessageListener
            for _, listener in self._listener_dicts[message_type].items():
                listener.func(message)

    @property
    def json(self):
        return json.dumps(self._listener_dicts, default=str)


messaging = Messaging()  # singleton



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240111

from random import random

from rich.console import Console
from rich.table import Table
from rich.text import Text

from neetbox.utils import Registry
from neetbox.utils.mvc import Singleton

from .._formatting import RawLog

LogWriters = Registry("LOG_WRITERS")


console = Console()
whom2color = {}

supported_colors = ["red", "green", "blue", "cyan", "yellow", "magenta"]
supported_text_styles = ["bold", "italic", "blink", "dim"]


@LogWriters.register(name="stdout")
def log_write_stdout(log: RawLog):
    series = (
        {
            "ok": Text("[ok]", style="green"),
            "debug": Text("[debug]", style="cyan"),
            "info": Text("[info]", style="white"),
            "warning": Text("[warning]", style="yellow"),
            "mention": Text("[mention]", style="yellow"),
            "error": Text("[error]", style="red"),
        }.get(log.series, Text(log.series))
        if log.series
        else Text("")
    )
    table = Table(show_header=False, box=None, expand=True)
    table.add_column(justify="left")
    table.add_column(justify="right")
    whom = log.caller_name_alias or log.caller_info_formatted
    if whom not in whom2color:
        whom2color[whom] = supported_colors[int(random() * len(supported_colors))]
    color = whom2color[whom]
    whom_text = Text(whom, color)
    split_text = Text(" > ", style="dim " + color)
    message_text = Text(log.message, style="default")
    time_text = Text(log.timestamp.strftime(r"%H:%M:%S"), style="default dim")
    table.add_row(
        whom_text + split_text + message_text, series + " " + time_text if series else time_text
    )
    console.print(table)



import numpy as np
import torch
import torch.nn as nn

from .kernels import *


class CannyFilter(nn.Module):
    def __init__(
        self,
        k_gaussian=3,
        mu=0,
        sigma=1,
        k_sobel=3,
        padding=True,
        use_cuda=False,
    ):
        super(CannyFilter, self).__init__()
        # device
        self.device = "cuda" if use_cuda else "cpu"
        # gaussian
        gaussian_2D = get_gaussian_kernel(k_gaussian, mu, sigma)
        # gaussian_2D = gaussian_2D[None, None, :]
        self.gaussian_filter = nn.Conv2d(
            in_channels=1,
            out_channels=1,
            kernel_size=k_gaussian,
            padding=k_gaussian // 2 if padding else 0,
            bias=False,
        )
        self.gaussian_filter.weight.data.copy_(torch.from_numpy(gaussian_2D[None, None, :]))

        # sobel
        sobel_2D = get_sobel_kernel(k_sobel)
        self.sobel_filter_x = nn.Conv2d(
            in_channels=1,
            out_channels=1,
            kernel_size=k_sobel,
            padding=k_sobel // 2 if padding else 0,
            bias=False,
        )
        self.sobel_filter_x.weight.data.copy_(torch.from_numpy(sobel_2D))

        self.sobel_filter_y = nn.Conv2d(
            in_channels=1,
            out_channels=1,
            kernel_size=k_sobel,
            padding=k_sobel // 2 if padding else 0,
            bias=False,
        )
        self.sobel_filter_y.weight.data.copy_(torch.from_numpy(sobel_2D.T))

        # thin

        thin_kernels = get_thin_kernels()
        directional_kernels = np.stack(thin_kernels)
        directional_kernels = np.expand_dims(directional_kernels, axis=1)
        self.directional_filter = nn.Conv2d(
            in_channels=1,
            out_channels=8,
            kernel_size=thin_kernels[0].shape,
            padding=thin_kernels[0].shape[-1] // 2 if padding else 0,
            bias=False,
        )
        self.directional_filter.weight.data.copy_(torch.from_numpy(directional_kernels))

        # hysteresis

        hysteresis = np.ones((3, 3)) + 0.25
        self.hysteresis = nn.Conv2d(
            in_channels=1, out_channels=1, kernel_size=3, padding=1, bias=False
        )
        self.hysteresis.weight.data.copy_(torch.from_numpy(hysteresis))
        self.hysteresis.weight.data.copy_(torch.from_numpy(hysteresis))

    @torch.no_grad()
    def forward(self, img, low_threshold=None, high_threshold=None, hysteresis=False):
        # set the setps tensors
        B, C, H, W = img.shape
        blurred = torch.zeros((B, C, H, W)).to(self.device)
        grad_x = torch.zeros((B, 1, H, W)).to(self.device)
        grad_y = torch.zeros((B, 1, H, W)).to(self.device)
        grad_magnitude = torch.zeros((B, 1, H, W)).to(self.device)
        grad_orientation = torch.zeros((B, 1, H, W)).to(self.device)

        # gaussian

        for c in range(C):
            blurred[:, c : c + 1] = self.gaussian_filter(img[:, c : c + 1])

            grad_x = grad_x + self.sobel_filter_x(blurred[:, c : c + 1])
            grad_y = grad_y + self.sobel_filter_y(blurred[:, c : c + 1])

        # thick edges

        grad_x, grad_y = grad_x / C, grad_y / C
        grad_magnitude = (grad_x**2 + grad_y**2) ** 0.5
        grad_orientation = torch.atan(grad_y / grad_x)
        grad_orientation = grad_orientation * (360 / np.pi) + 180  # convert to degree
        grad_orientation = torch.round(grad_orientation / 45) * 45  # keep a split by 45

        # thin edges

        directional = self.directional_filter(grad_magnitude)
        # get indices of positive and negative directions
        positive_idx = (grad_orientation / 45) % 8
        negative_idx = ((grad_orientation / 45) + 4) % 8
        thin_edges = grad_magnitude.clone()
        # non maximum suppression direction by direction
        for pos_i in range(4):
            neg_i = pos_i + 4
            # get the oriented grad for the angle
            is_oriented_i = (positive_idx == pos_i) * 1
            is_oriented_i = is_oriented_i + (positive_idx == neg_i) * 1
            pos_directional = directional[:, pos_i]
            neg_directional = directional[:, neg_i]
            selected_direction = torch.stack([pos_directional, neg_directional])

            # get the local maximum pixels for the angle
            is_max = selected_direction.min(dim=0)[0] > 0.0
            is_max = torch.unsqueeze(is_max, dim=1)

            # apply non-maximum suppression
            to_remove = (is_max == 0) * 1 * (is_oriented_i) > 0
            thin_edges[to_remove] = 0.0

        # thresholds

        if low_threshold is not None:
            low = thin_edges > low_threshold

            if high_threshold is not None:
                high = thin_edges > high_threshold
                # get black/gray/white only
                thin_edges = low * 0.5 + high * 0.5

                if hysteresis:
                    # get weaks and check if they are high or not
                    weak = (thin_edges == 0.5) * 1
                    weak_is_high = (self.hysteresis(thin_edges) > 1) * weak
                    thin_edges = high * 1 + weak_is_high * 1
            else:
                thin_edges = low * 1

        return blurred, grad_x, grad_y, grad_magnitude, grad_orientation, thin_edges



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231201

import collections
import json
import os
import sqlite3
from datetime import datetime
from threading import Lock
from typing import Union

from neetbox._protocol import *
from neetbox.config.user import get as get_global_config
from neetbox.logging import Logger
from neetbox.utils import ResourceLoader
from neetbox.utils.localstorage import get_file_size_in_bytes

from .._manager import manager
from ..abc import FetchType, ManageableDB, SortType
from .condition import ProjectDbQueryCondition

logger = Logger("PROJECT DB", skip_writers_names=["ws"])
DB_PROJECT_FILE_FOLDER = f"{get_global_config('vault')}/server/history"
DB_PROJECT_FILE_TYPE_NAME = "projectdb"


class ProjectDB(ManageableDB):
    # static things
    _path2dbc = {}

    # not static. instance level vars
    project_id: str  # of which project id
    file_path: str  # where is the db file
    connection: sqlite3.Connection  # the db connection
    _inited_tables: collections.defaultdict

    def __new__(cls, project_id: str = None, path: str = None, **kwargs) -> "ProjectDB":
        if path is None and project_id is None:
            raise RuntimeError(f"please provide at least project id or path when creating db")
        if path is None:  # make path from project id
            path = f"{DB_PROJECT_FILE_FOLDER}/{project_id}.{DB_PROJECT_FILE_TYPE_NAME}"
        if path in cls._path2dbc:
            return cls._path2dbc[path]
        if project_id in manager.current:
            return manager.current[project_id]
        new_dbc = super().__new__(cls, **kwargs)
        # connect to sqlite
        new_dbc.file_path = path
        new_dbc.connection = sqlite3.connect(path, check_same_thread=False, isolation_level=None)
        new_dbc.connection.execute("pragma journal_mode=wal")  # set journal mode WAL
        new_dbc.connection.execute("PRAGMA foreign_keys = ON")  # enable foreign keys features
        new_dbc._inited_tables = collections.defaultdict(lambda: False)
        # check neetbox version
        _db_file_project_id = new_dbc.fetch_db_project_id(project_id)
        project_id = project_id or _db_file_project_id
        if _db_file_project_id != project_id:
            raise RuntimeError(
                f"Wrong DB file! reading history of project id '{project_id}' from file of project id '{_db_file_project_id}'"
            )
        _db_file_version = new_dbc.fetch_db_version(NEETBOX_VERSION)
        if NEETBOX_VERSION != _db_file_version:
            logger.warn(
                f"History file version not match: reading from version {_db_file_version} with neetbox version {NEETBOX_VERSION}"
            )
        cls._path2dbc[path] = new_dbc
        manager.current[project_id] = new_dbc
        new_dbc.project_id = project_id
        logger.ok(f"History file(version={_db_file_version}) for project id '{project_id}' loaded.")
        return new_dbc

    @property
    def size(self):
        """return local storage usage by this project db in bytes

        Returns:
            int: size in bytes
        """
        try:
            result = get_file_size_in_bytes(self.file_path)
        except Exception as e:
            result = f"failed to get file size cause {e}"
        return result

    def close(self):
        try:
            self.connection.commit()
        except:
            pass
        self.connection.close()

    def delete(self):
        """delete related files of db"""
        if self.project_id not in manager.current:
            logger.err(
                RuntimeError(f"could not find db to delete with project id {self.project_id}")
            )
        del manager.current[self.project_id]
        del ProjectDB._path2dbc[self.file_path]
        logger.info(f"deleting history DB for project id {self.project_id}...")
        if self.connection:
            try:
                self.connection.close()
            except Exception as e:
                logger.err(
                    RuntimeError(
                        f"failed to close DB connection for project id {self.project_id} because {e}"
                    ),
                    reraise=True,
                )
        try:
            for suffix in ["", "-shm", "-wal"]:  # remove db files
                _path = f"{self.file_path}{suffix}"
                if os.path.exists(_path):
                    os.remove(_path)
        except Exception as e:
            logger.err(
                RuntimeError(
                    f"failed to delete DB file for project id {self.project_id} because {e}"
                ),
                reraise=True,
            )
        logger.info(f"History db for project id {self.project_id} has been deleted.")

    @classmethod
    def items(cls):
        return manager.current.items()

    @classmethod
    def of_project_id(cls, project_id):
        if project_id in manager.current:
            return manager.current[project_id]
        return ProjectDB(project_id)

    def _execute(self, query, *args, fetch: FetchType = FetchType.ALL, **kwargs):
        cur = self.connection.cursor()
        try:
            result = cur.execute(query, args)
        except Exception as e:
            logger.err(f"failed to execute query cause '{e}'")
            logger.info(f"{query}, {args}")
            logger.err(e, reraise=True)
        if fetch:
            if fetch == FetchType.ALL:
                result = result.fetchall()
            elif fetch == FetchType.ONE:
                result = result.fetchone()
            elif fetch == FetchType.MANY:
                result = result.fetchmany(kwargs["many"])
        return result, cur.lastrowid

    def _query(self, query, *args, fetch: FetchType = FetchType.ALL, **kwargs):
        return self._execute(query, *args, fetch=fetch, **kwargs)

    def table_exist(self, table_name):
        sql_query = f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';"
        result, _ = self._query(sql_query, fetch=FetchType.ALL)
        return result != []

    def get_table_names(self):
        sql_query = "SELECT name FROM sqlite_master;"
        table_names, _ = self._query(sql_query, fetch=FetchType.ALL)
        return table_names

    def fetch_db_version(self, default=None):
        if not self._inited_tables[VERSION_TABLE_NAME]:  # create if there is no version table
            sql_query = f"CREATE TABLE IF NOT EXISTS {VERSION_TABLE_NAME} ( {VERSION_TABLE_NAME} TEXT NON NULL );"
            self._execute(sql_query)
            self._inited_tables[VERSION_TABLE_NAME] = True
        sql_query = f"SELECT {VERSION_TABLE_NAME} FROM {VERSION_TABLE_NAME}"
        _version, _ = self._query(sql_query, fetch=FetchType.ONE)
        if _version is None:
            if default is None:
                raise RuntimeError(
                    "should provide a default version if fetching from empty version"
                )
            sql_query = f"INSERT INTO {VERSION_TABLE_NAME} VALUES (?);"
            self._execute(sql_query, default)
            return default
        return _version[0]

    def fetch_db_project_id(self, default=None):
        if not self._inited_tables[PROJECT_ID_TABLE_NAME]:  # create if there is no project id table
            sql_query = f"CREATE TABLE IF NOT EXISTS {PROJECT_ID_TABLE_NAME} ( {PROJECT_ID_TABLE_NAME} TEXT NON NULL );"
            self._execute(sql_query)
            self._inited_tables[PROJECT_ID_TABLE_NAME] = True
        sql_query = f"SELECT {PROJECT_ID_TABLE_NAME} FROM {PROJECT_ID_TABLE_NAME}"
        _projectid, _ = self._query(sql_query, fetch=FetchType.ONE)
        if _projectid is None:
            if default is None:
                raise RuntimeError(
                    "should provide a default project id if fetching from empty project id"
                )
            sql_query = f"INSERT INTO {PROJECT_ID_TABLE_NAME} VALUES (?);"
            self._execute(sql_query, default)
            return default
        return _projectid[0]

    def get_id_of_run_id(self, run_id: str):
        try:
            sql_query = f"SELECT {ID_COLUMN_NAME} FROM {RUN_IDS_TABLE_NAME} WHERE {RUN_ID_COLUMN_NAME} == '{run_id}'"
            _id, _ = self._query(sql_query, fetch=FetchType.ONE)
            return _id[0]
        except:
            return None

    _run_id_fetch_lock = Lock()

    def fetch_id_of_run_id(self, run_id: str, timestamp: str = None):
        if not self._inited_tables[RUN_IDS_TABLE_NAME]:  # create if there is no version table
            sql_query = f"CREATE TABLE IF NOT EXISTS {RUN_IDS_TABLE_NAME} ( {ID_COLUMN_NAME} INTEGER PRIMARY KEY AUTOINCREMENT, {RUN_ID_COLUMN_NAME} TEXT NON NULL, {TIMESTAMP_COLUMN_NAME} TEXT NON NULL, {METADATA_COLUMN_NAME} TEXT, CONSTRAINT run_id_unique UNIQUE ({RUN_ID_COLUMN_NAME}));"
            self._execute(sql_query)
            self._inited_tables[RUN_IDS_TABLE_NAME] = True
        with self._run_id_fetch_lock:
            id_of_run_id = self.get_id_of_run_id(run_id)
            if id_of_run_id is None:
                timestamp = timestamp or datetime.now().strftime(DATETIME_FORMAT)
                sql_query = f"INSERT INTO {RUN_IDS_TABLE_NAME}({RUN_ID_COLUMN_NAME}, {TIMESTAMP_COLUMN_NAME})   VALUES (?, ?)"
                _, lastrowid = self._execute(sql_query, run_id, timestamp)
                return lastrowid
        return id_of_run_id

    def fetch_metadata_of_run_id(self, run_id: str, metadata: Union[dict, str] = None):
        id_of_run_id = self.get_id_of_run_id(run_id)
        if id_of_run_id is None:
            return None
        if metadata:  # if update name
            metadata = json.dumps(metadata) if isinstance(metadata, dict) else metadata
            sql_query = f"UPDATE {RUN_IDS_TABLE_NAME} SET {METADATA_COLUMN_NAME} = ? WHERE {ID_COLUMN_NAME} = ?"
            _, _ = self._execute(sql_query, metadata, id_of_run_id)
        # get name
        sql_query = f"SELECT {METADATA_COLUMN_NAME} FROM {RUN_IDS_TABLE_NAME} WHERE {ID_COLUMN_NAME} == {id_of_run_id}"
        (metadata,), _ = self._query(sql_query, fetch=FetchType.ONE)
        metadata = (
            json.loads(metadata) if metadata else {}
        )  # if does not have metadata, return an empty one
        return metadata

    def get_run_id_of_id(self, id_of_run_id):
        try:
            sql_query = f"SELECT {RUN_ID_COLUMN_NAME} FROM {RUN_IDS_TABLE_NAME} WHERE {ID_COLUMN_NAME} == {id_of_run_id}"
            run_id, _ = self._query(sql_query, fetch=FetchType.ONE)
            return run_id[0]
        except:
            return None

    def get_run_ids(self):
        if not self.table_exist(RUN_IDS_TABLE_NAME):
            return []
        sql_query = f"SELECT {RUN_ID_COLUMN_NAME}, {TIMESTAMP_COLUMN_NAME}, {METADATA_COLUMN_NAME} FROM {RUN_IDS_TABLE_NAME}"
        result, _ = self._query(sql_query)
        result = [
            {
                RUN_ID_KEY: run_id,
                TIMESTAMP_KEY: timestamp,
                METADATA_KEY: json.loads(metadata) if metadata else {},
            }
            for run_id, timestamp, metadata in result
        ]
        return result

    def delete_run_id(self, run_id: str):
        sql_query = f"DELETE FROM {RUN_IDS_TABLE_NAME} where {RUN_ID_COLUMN_NAME} = ?"
        _, _ = self._execute(sql_query, run_id)

    def get_series_of_table(self, table_name, run_id=None):
        if not self.table_exist(table_name):
            return []
        if run_id is not None:
            sql_query = f"SELECT DISTINCT t.series as series FROM {RUN_IDS_TABLE_NAME} r LEFT JOIN {table_name} t ON r.{RUN_ID_COLUMN_NAME} == ? WHERE t.{RUN_ID_COLUMN_NAME} == r.{ID_COLUMN_NAME}"
            args = (run_id,)
        else:
            sql_query = f"SELECT DISTINCT series FROM {table_name}"
            args = ()
        result, _ = self._query(sql_query, *args, fetch=FetchType.ALL)
        return [result for (result,) in result]

    def do_limit_num_row_for(
        self, table_name: str, run_id: str, num_row_limit: int, series: str = None
    ):
        if num_row_limit <= 0:  # no limit or random not triggered
            return
        sql_query = f"SELECT count(*) from {table_name} WHERE {RUN_ID_COLUMN_NAME} = {run_id}"  # count rows for run id in specific table
        if series is not None:
            sql_query += f" AND {SERIES_COLUMN_NAME} = '{series}'"
        num_rows, _ = self._query(sql_query, fetch=FetchType.ONE)
        if num_rows[0] > num_row_limit:  # num rows exceeded limit
            sql_query = f"SELECT {ID_COLUMN_NAME} from {table_name} WHERE {RUN_ID_COLUMN_NAME} = {run_id} ORDER BY {ID_COLUMN_NAME} DESC LIMIT 1 OFFSET {num_row_limit - 1}"  # get max id to delete of row for specific run id
            max_id_to_del, _ = self._query(sql_query, fetch=FetchType.ONE)
            sql_query = f"DELETE FROM {table_name} WHERE {ID_COLUMN_NAME} < {max_id_to_del[0]} AND {RUN_ID_COLUMN_NAME} = {run_id}"
            if series is not None:
                sql_query += f" AND {SERIES_COLUMN_NAME} = '{series}'"
            self._query(sql_query)  # delete rows with smaller id and specific run id

    def write_json(
        self,
        table_name: str,
        json_data: str,
        series: str = None,
        run_id: str = None,
        timestamp: str = None,
        num_row_limit=-1,
    ):
        if not isinstance(json_data, dict):
            json_data = json.loads(json_data)
        if run_id:
            run_id = self.fetch_id_of_run_id(run_id, timestamp=timestamp)
        if not self._inited_tables[table_name]:  # create if there is no version table
            sql_query = f"CREATE TABLE IF NOT EXISTS {table_name} ( {ID_COLUMN_NAME} INTEGER PRIMARY KEY AUTOINCREMENT, {TIMESTAMP_COLUMN_NAME} TEXT NON NULL, {SERIES_COLUMN_NAME} TEXT, {RUN_ID_COLUMN_NAME} INTEGER, {JSON_COLUMN_NAME} TEXT NON NULL, FOREIGN KEY({RUN_ID_COLUMN_NAME}) REFERENCES {RUN_IDS_TABLE_NAME}({ID_COLUMN_NAME}) ON DELETE CASCADE);"
            self._execute(sql_query)
            sql_query = f"CREATE INDEX IF NOT EXISTS series_and_runid_index ON {table_name} ({SERIES_COLUMN_NAME}, {RUN_ID_COLUMN_NAME})"
            self._execute(sql_query)
            self._inited_tables[table_name] = True

        sql_query = f"INSERT INTO {table_name}({TIMESTAMP_COLUMN_NAME}, {SERIES_COLUMN_NAME}, {RUN_ID_COLUMN_NAME}, {JSON_COLUMN_NAME}) VALUES (?, ?, ?, ?)"
        if isinstance(json_data, dict):
            json_data = json.dumps(json_data)
        _, lastrowid = self._execute(sql_query, timestamp, series, run_id, json_data)
        self.do_limit_num_row_for(
            table_name=table_name, run_id=run_id, num_row_limit=num_row_limit, series=series
        )
        return lastrowid

    def read_json(self, table_name: str, condition: ProjectDbQueryCondition = None):
        if not self.table_exist(table_name):
            return []
        if condition and isinstance(condition.run_id, str):
            condition.run_id = self.get_id_of_run_id(condition.run_id)  # convert run id
        cond_str, cond_vars = condition.dumpt() if condition else ""
        sql_query = f"SELECT {', '.join((ID_COLUMN_NAME, TIMESTAMP_COLUMN_NAME,SERIES_COLUMN_NAME, JSON_COLUMN_NAME))} FROM {table_name} {cond_str}"
        result, _ = self._query(sql_query, *cond_vars, fetch=FetchType.ALL)
        result = [
            {
                ID_COLUMN_NAME: w,
                TIMESTAMP_COLUMN_NAME: x,
                SERIES_COLUMN_NAME: y,
                JSON_COLUMN_NAME: json.loads(z) if z else None,
            }
            for w, x, y, z in result
        ]
        return result

    def set_status(self, run_id: str, series: str, json_data):
        if not (isinstance(json_data, str) or isinstance(json_data, dict)):
            raise
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        if run_id:
            run_id = self.fetch_id_of_run_id(run_id)
        if not self._inited_tables[STATUS_TABLE_NAME]:  # create if there is no version table
            sql_query = f"CREATE TABLE IF NOT EXISTS {STATUS_TABLE_NAME} ({ID_COLUMN_NAME} INTEGER PRIMARY KEY,{RUN_ID_COLUMN_NAME} INTEGER NON NULL, {SERIES_COLUMN_NAME} TEXT NON NULL, {JSON_COLUMN_NAME} TEXT NON NULL, UNIQUE({RUN_ID_COLUMN_NAME}, {SERIES_COLUMN_NAME}) ON CONFLICT REPLACE, FOREIGN KEY({RUN_ID_COLUMN_NAME}) REFERENCES {RUN_IDS_TABLE_NAME}({ID_COLUMN_NAME}) ON DELETE CASCADE);"
            self._execute(sql_query)
            self._inited_tables[STATUS_TABLE_NAME] = True
        sql_query = f"INSERT OR REPLACE INTO {STATUS_TABLE_NAME}({RUN_ID_COLUMN_NAME},{SERIES_COLUMN_NAME}, {JSON_COLUMN_NAME}) VALUES (?, ?, ?)"
        if isinstance(json_data, dict):
            json_data = json.dumps(json_data)
        _, lastrowid = self._execute(sql_query, run_id, series, json_data)
        return lastrowid

    def get_status(self, run_id: str = None, series: str = None):
        if not self.table_exist(STATUS_TABLE_NAME):
            return {}
        condition = ProjectDbQueryCondition(run_id=run_id, series=series)
        if isinstance(condition.run_id, str):
            condition.run_id = self.get_id_of_run_id(run_id)
        cond_str, cond_vars = condition.dumpt()
        sql_query = f"SELECT {', '.join((RUN_ID_COLUMN_NAME, SERIES_COLUMN_NAME, JSON_COLUMN_NAME))} FROM {STATUS_TABLE_NAME} {cond_str}"
        query_result, _ = self._query(sql_query, *cond_vars, fetch=FetchType.ALL)
        result = {}
        for id_of_runid, series_name, value in query_result:
            run_id = self.get_run_id_of_id(id_of_runid)
            if run_id and run_id not in result:
                result[run_id] = {}
            result[run_id][series_name] = json.loads(value)
        return result

    def write_blob(
        self,
        table_name: str,
        meta_data: Union[str, dict],
        blob_data: bytes,
        series: str = None,
        run_id: str = None,
        timestamp: str = None,
        num_row_limit=-1,
    ):
        meta_data = meta_data or {}
        meta_data = meta_data if isinstance(meta_data, dict) else json.loads(meta_data)
        if run_id:
            run_id = self.fetch_id_of_run_id(run_id, timestamp=timestamp)
        if isinstance(blob_data, bytes):
            blob_data = bytearray(blob_data)
        if isinstance(meta_data, dict):
            meta_data = json.dumps(meta_data)

        if not self._inited_tables[table_name]:  # create if not exist
            sql_query = f"CREATE TABLE IF NOT EXISTS {table_name} ( {ID_COLUMN_NAME} INTEGER PRIMARY KEY AUTOINCREMENT, {TIMESTAMP_COLUMN_NAME} TEXT NON NULL, {SERIES_COLUMN_NAME} TEXT, {RUN_ID_COLUMN_NAME} INTEGER, {METADATA_COLUMN_NAME} TEXT, {BLOB_COLUMN_NAME} BLOB NON NULL, FOREIGN KEY({RUN_ID_COLUMN_NAME}) REFERENCES {RUN_IDS_TABLE_NAME}({ID_COLUMN_NAME}) ON DELETE CASCADE);"
            self._execute(sql_query)
            sql_query = f"CREATE INDEX IF NOT EXISTS series_and_runid_index ON {table_name} ({SERIES_COLUMN_NAME}, {RUN_ID_COLUMN_NAME})"
            self._execute(sql_query)
            self._inited_tables[table_name] = True

        sql_query = f"INSERT INTO {table_name}({TIMESTAMP_COLUMN_NAME}, {SERIES_COLUMN_NAME}, {RUN_ID_COLUMN_NAME}, {METADATA_COLUMN_NAME}, {BLOB_COLUMN_NAME}) VALUES (?, ?, ?, ?, ?)"
        _, lastrowid = self._execute(sql_query, timestamp, series, run_id, meta_data, blob_data)
        self.do_limit_num_row_for(
            table_name=table_name, run_id=run_id, num_row_limit=num_row_limit, series=series
        )
        return lastrowid

    def read_blob(
        self, table_name: str, condition: ProjectDbQueryCondition = None, meta_only=False
    ):
        if not self.table_exist(table_name):
            return []
        if condition and isinstance(condition.run_id, str):
            condition.run_id = self.get_id_of_run_id(condition.run_id)  # convert run id
        cond_str, cond_vars = condition.dumpt() if condition else ""
        sql_query = f"SELECT {', '.join((ID_COLUMN_NAME,TIMESTAMP_COLUMN_NAME, METADATA_COLUMN_NAME, *((BLOB_COLUMN_NAME,) if not meta_only else ())))} FROM {table_name} {cond_str}"
        result, _ = self._query(sql_query, *cond_vars, fetch=FetchType.ALL)
        return result

    @classmethod
    def load_db_of_path(cls, path):
        if not os.path.isfile(path):
            raise RuntimeError(f"{path} is not a file")
        conn = ProjectDB(path=path)
        return conn

    @classmethod
    def get_db_list(cls):
        history_file_loader = ResourceLoader(
            folder=DB_PROJECT_FILE_FOLDER, file_types=[DB_PROJECT_FILE_TYPE_NAME], force_rescan=True
        )
        history_file_list = history_file_loader.get_file_list()
        for path in history_file_list:
            cls.load_db_of_path(path=path)
        return manager.current.items()

    @classmethod
    def get_db_of_id(cls, project_id, rescan: bool = True):
        if rescan:
            cls.get_db_list()  # scan for possible file changes
        conn = ProjectDB.of_project_id(project_id=project_id)
        return conn


# === SCAN FOR DB FILES ===

if not os.path.exists(DB_PROJECT_FILE_FOLDER):
    # create history root dir
    logger.info(f"history file directory not exist, trying to create at {DB_PROJECT_FILE_FOLDER}")
    os.makedirs(DB_PROJECT_FILE_FOLDER)
# check if is dir
assert os.path.isdir(DB_PROJECT_FILE_FOLDER), f"{DB_PROJECT_FILE_FOLDER} is not a directory."
logger.info(f"using history file folder: {DB_PROJECT_FILE_FOLDER}")



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240111

import os

from .._formatting import RawLog


class FileLogWriter:
    # class level static
    PATH2WRITER = {}

    # instance level
    file_writer = None

    def __new__(cls, path):
        path = os.path.abspath(path)
        if os.path.isdir(path):
            raise Exception("Target path is not a file.")
        dirname = os.path.dirname(path) if len(os.path.dirname(path)) != 0 else "."
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        if path in FileLogWriter.PATH2WRITER:
            return FileLogWriter.PATH2WRITER[path]
        new_instance = super().__new__(cls)
        new_instance.file_writer = open(path, "a", encoding="utf-8", buffering=1)
        FileLogWriter.PATH2WRITER[path] = new_instance
        return new_instance

    def write(self, log: RawLog):
        line = " ".join(
            [
                log.timestamp.strftime(r"%Y-%m-%dT%H:%M:%S.%f"),
                f"[{log.series}]" if log.series else "[log]",
                log.caller_info.format(r"%m.%c.%f:%l"),
                str(log.message),
            ]
        )
        self.file_writer.write(line + "\n")



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231201

import json
from dataclasses import dataclass
from datetime import datetime as dt
from enum import Enum
from importlib.metadata import version
from typing import Any, Union

VERSION = version("neetbox")

# ===================== common things =====================

ID_KEY = "id"
NAME_KEY = "name"
ARGS_KEY = "args"
MACHINE_ID_KEY = "machineId"
PROJECT_KEY = "project"
SERVER_KEY = "server"
PROJECT_ID_KEY = "projectId"
RUN_ID_KEY = "runId"
SERIES_KEY = "series"
EVENT_TYPE_KEY = "eventType"
EVENT_ID_KEY = "eventId"
IDENTITY_TYPE_KEY = "identityType"
CALLER_ID_KEY = "whom"
PAYLOAD_KEY = "payload"
METADATA_KEY = "metadata"
MESSAGE_KEY = "message"
TIMESTAMP_KEY = "timestamp"
HISTORY_LEN_KEY = "historyLen"
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"  # YYYY-MM-DDTHH:MM:SS.SSS


def get_timestamp(datetime=None):
    datetime = datetime or dt.now()
    return datetime.strftime(DATETIME_FORMAT)


class IdentityType(str, Enum):
    WEB = "web"
    CLI = "cli"
    SERVER = "server"
    SELF = "self"
    OTHERS = "another"
    BOTH = "both"

    def __repr__(self) -> str:
        return self.value


RESULT_KEY = "result"
ERROR_KEY = "error"
REASON_KEY = "reason"


@dataclass
class EventMsg:
    project_id: str
    run_id: str
    event_type: str
    series: str = None
    payload: Any = None
    event_id: int = -1
    identity_type: str = None
    timestamp: str = get_timestamp()
    history_len: int = -1
    id: int = None  # id in database

    @property
    def json(self):
        return {
            PROJECT_ID_KEY: self.project_id,
            RUN_ID_KEY: self.run_id,
            EVENT_TYPE_KEY: self.event_type,
            EVENT_ID_KEY: self.event_id,
            IDENTITY_TYPE_KEY: self.identity_type,
            SERIES_KEY: self.series,
            PAYLOAD_KEY: self.payload,
            TIMESTAMP_KEY: self.timestamp,
            HISTORY_LEN_KEY: self.history_len,
            ID_KEY: self.id,
        }

    def dumps(self):
        return json.dumps(self.json, default=str)

    @classmethod
    def loads(cls, src):
        if isinstance(src, str):
            src = json.loads(src)
        return EventMsg(
            project_id=src.get(PROJECT_ID_KEY),
            run_id=src.get(RUN_ID_KEY),
            event_type=src.get(EVENT_TYPE_KEY),
            identity_type=src.get(IDENTITY_TYPE_KEY),
            series=src.get(SERIES_KEY),
            payload=src.get(PAYLOAD_KEY),
            event_id=src.get(EVENT_ID_KEY, -1),
            timestamp=src.get(TIMESTAMP_KEY, get_timestamp()),
            history_len=src.get(HISTORY_LEN_KEY, -1),
            id=src.get(ID_KEY, None),
        )

    @classmethod
    def merge(cls, x: Union["EventMsg", dict], y: Union["EventMsg", dict]):
        _x = x if isinstance(x, dict) else x.json
        _y = y if isinstance(y, dict) else y.json
        for _k, _v in _y.items():
            _x[_k] = _v
        return cls.loads(_x)


# ===================== WS things =====================

# known event names
EVENT_TYPE_NAME_HANDSHAKE = "handshake"
EVENT_TYPE_NAME_WAVEHANDS = "wavehands"
EVENT_TYPE_NAME_LOG = "log"
EVENT_TYPE_NAME_ACTION = "action"
EVENT_TYPE_NAME_SCALAR = "scalar"
EVENT_TYPE_NAME_HIST = "histogram"
EVENT_TYPE_NAME_IMAGE = "image"
EVENT_TYPE_NAME_VIDEO = "video"
EVENT_TYPE_NAME_HPARAMS = "hyperparameters"
EVENT_TYPE_NAME_STATUS = "status"
EVENT_TYPE_NAME_HARDWARE = "hardware"
EVENT_TYPE_NAME_PROGRESS = "progress"

# ===================== HTTP things =====================

API_ROOT = "/api"
WS_ROOT = "/ws"

# ===================== DB things =====================

# === COLUMN NAMES ===
ID_COLUMN_NAME = ID_KEY
TIMESTAMP_COLUMN_NAME = TIMESTAMP_KEY
NAME_COLUMN_NAME = SERIES_COLUMN_NAME = SERIES_KEY
RUN_ID_COLUMN_NAME = RUN_ID_KEY
JSON_COLUMN_NAME = METADATA_COLUMN_NAME = METADATA_KEY
BLOB_COLUMN_NAME = "data"

# === TABLE NAMES ===
PROJECT_ID_TABLE_NAME = PROJECT_ID_KEY
VERSION_TABLE_NAME = "version"
RUN_IDS_TABLE_NAME = RUN_ID_KEY
STATUS_TABLE_NAME = EVENT_TYPE_NAME_STATUS
LOG_TABLE_NAME = "log"
IMAGE_TABLE_NAME = "image"

NEETBOX_VERSION = version("neetbox")



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231206

from ._client import connection
from ._signal_and_slot import listen, watch
from .apis._action import actionManager
from .apis._image import add_figure, add_image
from .apis._progress import Progress as progress
from .apis._scalar import add_hyperparams, add_scalar

ws_subscribe = connection.ws_subscribe
action = actionManager.register

__all__ = [
    "add_image",
    "add_scalar",
    "add_figure",
    "add_hyperparams",
    "ws_subscribe",
    "action",
    "watch",
    "listen",
    "progress",
]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230315

import torch.nn as nn


def ConvNxN(
    inplanes,
    outplanes,
    kernel_size=3,
    stride=1,
    padding=True,
    dilation=1,
    depthwise=False,
):
    """ConvNxN

    Args:
        inplanes (int): num of channel input
        outplanes (int): num of channel output
        kernel_size (int, optional): kernel size. Defaults to 3.
        stride (int, optional): stride. Defaults to 1.
        padding (bool, optional): padding. Defaults to True.
        dilation (int, optional): dilation rate. Defaults to 1.

    Returns:
        nn.Conv2d: the convolution layer
    """
    if depthwise:
        assert outplanes % inplanes == 0
    padding_size = (kernel_size // 2) * dilation if padding else 0
    return nn.Conv2d(
        in_channels=inplanes,
        out_channels=outplanes,
        kernel_size=kernel_size,
        padding=(padding_size, padding_size),
        stride=stride,
        dilation=dilation,
        groups=1 if not depthwise else inplanes,
    )


def SpatialSeparableConvNxN(planes, kernel_size=3, padding=True, dilation=1, depthwise=False):
    """SpatialSeparableConvNxN

    Args:
        planes (int): num of channel input and output
        kernel_size (int, optional): kernel size. Defaults to 3.
        dilation (int, optional): dilation rate. Defaults to 1.

    Returns:
        _type_: _description_
    """
    padding_size = (kernel_size // 2) * dilation if padding else 0
    return nn.Sequential(
        nn.Conv2d(
            in_channels=planes,
            out_channels=planes,
            kernel_size=(kernel_size, 1),
            padding=(padding_size, 0),
            stride=1,
            dilation=dilation,
            groups=1 if not depthwise else planes,
        ),
        nn.Conv2d(
            in_channels=planes,
            out_channels=planes,
            kernel_size=(1, kernel_size),
            padding=(0, padding_size),
            stride=1,
            dilation=dilation,
            groups=1 if not depthwise else planes,
        ),
    )


class ResBlock(nn.Module):
    def __init__(
        self,
        inplanes,
        outplanes,
        kernel_size=3,
        stride=1,
        padding=True,
        residual=False,
        spatial_separable=False,
        dilation=1,
        depthwise=False,
        pool_on_residual_downsample=False,
        bn_momentum=0.1,
        skip_last_relu=False,
    ):
        """ResBlock

        Args:
            inplanes (int): num of channel input
            outplanes (int): num of channel output
            kernel_size (int, optional): kernel size. Defaults to 3.
            stride (int, optional): stride for downsampling layer. Defaults to 1.
            padding (bool, optional): decide if use padding. Defaults to True.
            residual (bool, optional): wether use residual. Defaults to False.
            spatial_separable (bool, optional): set spatial separable for non-downsamping layers. Defaults to False.
            dilation (int, optional): dilation rate. Defaults to 1.
            depthwise (bool, optional): wether to use depthwise convolution. Defaults to False.
            pool_on_residual_downsample (bool, optional): 'maxpool' or 'averagepool' if you want to use pooling instead of conv2d on residual path. Defaults to False.
            bn_momentum (float, optional): momentum of batch norms. Defaults to 0.1.
            skip_last_relu (bool, optional): wether to skip the last relu. Defaults to False.
        """
        super(ResBlock, self).__init__()
        residual_padding_size = kernel_size // 2 if padding else 0
        self.conv1 = ConvNxN(
            inplanes=inplanes,
            outplanes=outplanes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            depthwise=depthwise,
        )
        self.bn1 = nn.BatchNorm2d(num_features=outplanes, momentum=bn_momentum)
        self.relu_inplace = nn.ReLU(inplace=True)
        if not spatial_separable:
            self.conv2 = ConvNxN(
                inplanes=outplanes,
                outplanes=outplanes,
                kernel_size=kernel_size,
                stride=1,
                padding=padding,
                dilation=dilation,
                depthwise=depthwise,
            )
        else:
            self.conv2 = SpatialSeparableConvNxN(
                planes=outplanes,
                kernel_size=kernel_size,
                dilation=dilation,
                depthwise=depthwise,
            )
        self.bn2 = nn.BatchNorm2d(num_features=outplanes, momentum=bn_momentum)
        self.residual = None
        if residual:
            if pool_on_residual_downsample:
                assert inplanes == outplanes
                assert pool_on_residual_downsample in ["maxpool", "averagepool"]
                if pool_on_residual_downsample == "maxpool":
                    self.residual = nn.MaxPool2d(
                        kernel_size=kernel_size,
                        stride=stride,
                        padding=residual_padding_size,
                    )
                elif pool_on_residual_downsample == "averagepool":
                    self.residual = nn.AvgPool2d(
                        kernel_size=kernel_size,
                        stride=stride,
                        padding=residual_padding_size,
                    )
            else:
                self.residual = nn.Sequential(
                    ConvNxN(
                        inplanes=inplanes,
                        outplanes=outplanes,
                        kernel_size=1,
                        stride=stride,
                    ),
                    nn.BatchNorm2d(num_features=outplanes, momentum=bn_momentum),
                )
        self.skip_last_relu = skip_last_relu

    def forward(self, x):
        _x = self.conv1(x)
        _x = self.bn1(_x)
        _x = self.relu_inplace(_x)
        _x = self.conv2(_x)
        _x = self.bn2(_x)
        if self.residual:
            _x_res = self.residual(x)
            _x = _x + _x_res
        if not self.skip_last_relu:
            _x = self.relu_inplace(_x)
        return _x


class InceptionBlock(nn.Module):
    def __init__(self, **kwargs) -> None:
        super(InceptionBlock, self).__init__()



from typing import Union

import torch


def one_hot(
    tensor: torch.Tensor,
    num_classes: int,
    ignored_label: Union[str, int] = "negative",
):
    """An advanced version of F.one_hot with ignore label support. Convert the mask to a one-hot encoded representation by @visualDust

    Args:
        tensor (torch.Tensor): indexed label image. Should types int
        num_classes (int): number of classes
        ignored_label (Union[str|int], optional): specify labels to ignore, or ignore by pattern. Defaults to "negative".

    Returns:
        torch.Tensor: one hot encoded tensor
    """
    original_shape = tensor.shape
    for _ in range(4 - len(tensor.shape)):
        tensor = tensor.unsqueeze(0)  # H W -> C H W -> B C H W, if applicable
    # start to handle ignored label
    # convert ignored label into positive index bigger than num_classes
    if type(ignored_label) is int:
        tensor[tensor == ignored_label] = num_classes
    elif ignored_label == "negative":
        tensor[tensor < 0] = num_classes

    # check if mask image is valid
    if torch.max(tensor) > num_classes:
        raise RuntimeError("class values must be smaller than num_classes.")
    B, _, H, W = tensor.shape
    one_hot = torch.zeros(B, num_classes + 1, H, W)
    one_hot.scatter_(1, tensor, 1)  # mark 1 on channel(dim=1) with index of mask
    one_hot = one_hot[:, :num_classes]  # remove ignored label(s)
    for _ in range(len(one_hot.shape) - len(original_shape)):
        one_hot.squeeze_(0)  # B C H W -> H W ->  C H W, if applicable
    return one_hot



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230413


import inspect
import os
import sys
import types
import uuid
from importlib.metadata import version
from threading import Thread
from typing import Union
from uuid import uuid4

import toml

from neetbox._protocol import NAME_KEY, PROJECT_ID_KEY, RUN_ID_KEY
from neetbox.utils.framing import get_frame_module_traceback
from neetbox.utils.massive import check_read_toml, update_dict_recursively

CONFIG_FILE_NAME = f"neetbox.toml"
NEETBOX_VERSION = version("neetbox")

_DEFAULT_WORKSPACE_CONFIG = {
    NAME_KEY: os.path.basename(os.path.normpath(os.getcwd())),
    "version": NEETBOX_VERSION,
    PROJECT_ID_KEY: str(uuid4()),  # later will be overwrite by workspace config file
    "extension": {
        "autoload": True,
    },
    "client": {
        "enable": True,
        "host": "127.0.0.1",
        "port": 20202,
        "allowIpython": False,
        "mute": True,
        "mode": "detached",
        "uploadInterval": 1,
        "shell": {"enable": True, "daemon": True},
    },
}


def _obtain_new_run_id():
    """put a new run id into config. do not run before workspace loaded as a project.

    Returns:
        str: new run id
    """
    global _DEFAULT_WORKSPACE_CONFIG
    _DEFAULT_WORKSPACE_CONFIG[RUN_ID_KEY] = str(uuid4())
    return _DEFAULT_WORKSPACE_CONFIG[RUN_ID_KEY]


_QUERY_ADD_EXTENSION_DEFAULT_CONFIG = []

_IS_WORKSPACE_LOADED = False
_QUERY_AFTER_CONFIG_LOAD = []


def on_config_loaded(func):
    if _IS_WORKSPACE_LOADED:  # if workspace already loaded
        Thread(target=func, daemon=True).start()  # run target
    else:  # not loaded yet, append to query
        global _QUERY_AFTER_CONFIG_LOAD
        _QUERY_AFTER_CONFIG_LOAD.append(func)
    return func


def export_default_config(func):
    global _QUERY_ADD_EXTENSION_DEFAULT_CONFIG
    if _IS_WORKSPACE_LOADED:
        raise RuntimeError("should not add default config after workspace loaded")
    _QUERY_ADD_EXTENSION_DEFAULT_CONFIG.append(func)
    return func


def _build_global_config_dict_for_module(module, local_config):
    """build a global config from a local config of a module. for example:
    - local config in neetbox/moduleA/some.py: { { "a" : 1 } }
    - the returned global config: {
        moduleA: {
            some: {
                "a" : 1
            }
        }
    }

    Args:
        module (_type_): _description_
        local_config (_type_): _description_
    """

    def _form_global_config_dict_from_module_name_list(
        pass_in_a_empty_dict_to_operate, module_name_list, local_config
    ):
        """turn a local config to a global config.
        Args:
            pass_in_a_empty_dict_to_operate (dict): pass in a empty dict, this function directly write into the dict, no value is returned.
            module_name_list (list): list of splited module names, for example, should be ["neetbox", "moduleA", "some"] for module neetbox.moduleA.some.
            local_config (dict): the local config of the sub module
        """
        for module_name in module_name_list:
            pass_in_a_empty_dict_to_operate[module_name] = {}
            pass_in_a_empty_dict_to_operate = pass_in_a_empty_dict_to_operate[module_name]
        for _k, _v in local_config.items():
            pass_in_a_empty_dict_to_operate[_k] = _v

    try:
        config = {}
        module_names = module.__name__.split(".")
        if module_names[0] == "neetbox":
            module_names.pop(0)
        _form_global_config_dict_from_module_name_list(config, module_names, local_config)
        return config
    except Exception as e:
        raise e


def _update_default_config_from_config_register():
    """iterate through config register to read their default config and write the result into global config. DO NOT run after workspace loaded.

    Raises:
        e: any possible exception
    """
    global _DEFAULT_WORKSPACE_CONFIG
    for fun in _QUERY_ADD_EXTENSION_DEFAULT_CONFIG:
        try:
            local_config = fun()
            parsed_local_config = _build_global_config_dict_for_module(
                inspect.getmodule(fun), local_config
            )
            update_dict_recursively(_DEFAULT_WORKSPACE_CONFIG, parsed_local_config)
        except Exception as e:
            raise e


def _update_default_workspace_config_with(cfg: dict):
    global _DEFAULT_WORKSPACE_CONFIG
    update_dict_recursively(_DEFAULT_WORKSPACE_CONFIG, cfg)
    return _DEFAULT_WORKSPACE_CONFIG


def _init_workspace(path=None, **kwargs) -> bool:
    if path:
        os.chdir(path=path)
    current_path = os.getcwd()
    config_file_path = os.path.join(current_path, CONFIG_FILE_NAME)
    if not os.path.exists(config_file_path):  # config file not exist
        try:  # creating config file using default config
            with open(config_file_path, "w+") as config_file:
                import neetbox.extension as extension

                extension._scan_sub_modules()
                _update_default_config_from_config_register()  # load custom config into default config
                _config = _DEFAULT_WORKSPACE_CONFIG
                if NAME_KEY in kwargs and kwargs[NAME_KEY]:  # using given name
                    _config[NAME_KEY] = kwargs[NAME_KEY]
                else:  # using the folder name
                    _config[NAME_KEY] = os.path.basename(os.path.normpath(os.getcwd()))
                _config[PROJECT_ID_KEY] = str(uuid.uuid4())
                toml.dump(_config, config_file)
            return True
        except Exception as e:
            raise e
    else:  # config file exist:
        raise RuntimeError(f"{config_file_path} already exist")


def _load_workspace_config(folder=".", load_only=False):
    global _IS_WORKSPACE_LOADED
    config_from_file = os.path.join(folder, CONFIG_FILE_NAME)
    config_from_file = check_read_toml(config_from_file)  # check if config file is valid
    if not config_from_file:  # failed to load workspace config, exiting
        raise RuntimeError(f"Config file not exists in '{folder}'")
    if not load_only:
        import neetbox.extension as extension

        extension._scan_sub_modules()
        _update_default_config_from_config_register()  # load custom config into default config
        _obtain_new_run_id()  # obtain new run id

    if "version" not in config_from_file or config_from_file["version"] != NEETBOX_VERSION:
        print(
            RuntimeError(
                f"config file version not match: using neetbox version {NEETBOX_VERSION} but got config file version {config_from_file['version']}. Please delete neetbox.toml and recreate by 'neet init'"
            )
        )
        os._exit(-1)
    _update_default_workspace_config_with(config_from_file)  # load config file in

    if load_only:
        return

    if (
        len(sys.argv) > 0
        and sys.argv[0].endswith("neet")
        or ("NEETBOX_DAEMON_PROCESS" in os.environ and os.environ["NEETBOX_DAEMON_PROCESS"] == "1")
    ):
        pass
    else:  # on workspace loaded
        _IS_WORKSPACE_LOADED = True
        for func in _QUERY_AFTER_CONFIG_LOAD:
            Thread(target=func, daemon=True).start()


def _create_load_workspace(folder="."):
    is_workspace = check_read_toml(path=os.path.join(folder, CONFIG_FILE_NAME))
    if not is_workspace:
        _init_workspace(folder)
    _load_workspace_config()


def _get_module_level_config(module: Union[str, types.ModuleType] = None, **kwargs):
    """get a module level config from global config

    Args:
        module (str or module, optional): which module's config to get. Defaults to None(which means neetbox will automatically find the module in which this function is called). if you want to get all the global config, pass "@" for module.

    Returns:
        dict: the config you want.
    """
    try:
        stack_offset = 2 if "stack_offset" not in kwargs else kwargs["stack_offset"]
        module = (
            module or get_frame_module_traceback(stack_offset=stack_offset).__name__  # type: ignore
        )  # try to trace if module not given
        if type(module) is not str:  # try to trace the belonging module of the given object
            module = inspect.getmodule(module).__name__  # type: ignore
    except Exception:
        module = "@"  # faild to trace, returning all configs
    the_config = _DEFAULT_WORKSPACE_CONFIG
    sub_module_names = module.split(".")
    if len(sub_module_names):
        if sub_module_names[0] == "neetbox":
            # skip 'neetbox'
            sub_module_names.pop(0)
        for sub_module in sub_module_names:
            sub_module = sub_module.removeprefix("_").removesuffix("_")
            if sub_module not in the_config:
                return the_config
            the_config = the_config[sub_module]
    return the_config
